@@ -520,9 +520,14 @@
 	{0x03EF, 0x03EE - 0x03EF, 0},
 	{0x03F0, 0x039A - 0x03F0, 0},
 	{0x03F1, 0x03A1 - 0x03F1, 0},
-	{0x03F2, 0x03A3 - 0x03F2, 0},
+	{0x03F2, 0x03F9 - 0x03F2, 0},
 	{0x03F4, 0, 0x03B8 - 0x03F4},
 	{0x03F5, 0x0395 - 0x03F5, 0},
+	{0x03F7, 0, 0x03F8 - 0x03F7},
+	{0x03F8, 0x03F7 - 0x03F8, 0},
+	{0x03F9, 0, 0x03F2 - 0x03F9},
+	{0x03FA, 0, 0x03FB - 0x03FA},
+	{0x03FB, 0x03FA - 0x03FB, 0},
 	{0x0400, 0, 0x0450 - 0x0400},
 	{0x0401, 0, 0x0451 - 0x0401},
 	{0x0402, 0, 0x0452 - 0x0402},
@@ -891,6 +896,8 @@
 	{0x10423, 0, 0x1044B - 0x10423},
 	{0x10424, 0, 0x1044C - 0x10424},
 	{0x10425, 0, 0x1044D - 0x10425},
+	{0x10426, 0, 0x1044E - 0x10426},
+	{0x10427, 0, 0x1044F - 0x10427},
 	{0x10428, 0x10400 - 0x10428, 0},
 	{0x10429, 0x10401 - 0x10429, 0},
 	{0x1042A, 0x10402 - 0x1042A, 0},
@@ -929,6 +936,8 @@
 	{0x1044B, 0x10423 - 0x1044B, 0},
 	{0x1044C, 0x10424 - 0x1044C, 0},
 	{0x1044D, 0x10425 - 0x1044D, 0},
+	{0x1044E, 0x10426 - 0x1044E, 0},
+	{0x1044F, 0x10427 - 0x1044F, 0},
 	{0x1E00, 0, 0x1E01 - 0x1E00},
 	{0x1E01, 0x1E00 - 0x1E01, 0},
 	{0x1E02, 0, 0x1E03 - 0x1E02},
