@@ -100,7 +100,7 @@ mousegetch ()
 		dx: line * 8 (first is 0)
 	*/
 		if (mregs.x.bx == stat) {
-			if (in_menu_mouse_mode == True) {
+			if (in_menu_mouse_mode) {
 				new_button = movebutton;
 			} else {
 				return mousegetch ();
