@@ -7,7 +7,10 @@
    the terminal is configured to be in UTF-8 mode.
  */
 
-#include <stdlib.h>
+#include <string.h>
+
+extern char * getenv ();
+
 
 static
 char *
@@ -18,17 +21,17 @@ locale_encoding (var1, var2, var3)
 {
   char * encoding;
   char * region = (char *) 0;
-  char * locale = (char *) getenv (var1);
+  char * locale = getenv (var1);
   if (! locale || ! strchr (locale, '.')) {
     if (locale) {
 	region = (char *) strrchr (locale, '_');
     }
-    locale = (char *) getenv (var2);
+    locale = getenv (var2);
     if (! locale || ! strchr (locale, '.')) {
 	if (locale && ! region) {
 		region = (char *) strrchr (locale, '_');
 	}
-	locale = (char *) getenv (var3);
+	locale = getenv (var3);
 	if (locale && ! region) {
 		region = (char *) strrchr (locale, '_');
 	}
