@@ -6,26 +6,44 @@
 #include "mined.h"
 #include "io.h"
 
+
 #ifndef __TURBOC__
 #include <sys/time.h>
 #endif
 #ifdef CURSES
 #include <curses.h>
+#include <strings.h>
 #endif
 
 
 voidfunc keyproc = I;	/* function addressed by entered function key */
+#define max_ansi_params 10
+int ansi_params;
+int ansi_param [max_ansi_params];
+char ansi_fini;
 char keyshift = '0';	/* shift state of entered function key */
 
 static
-int keymap_delay = 900;		/* wait to detect mapped key sequence */
+int keymap_delay = 0;		/* wait to detect mapped key sequence */
+static
+int default_keymap_delay = 900;	/* overridden by $MAPDELAY */
 
 static
-int escape_delay = 450;		/* wait to detect escape sequence */
+int escape_delay = 0;		/* wait to detect escape sequence */
+static
+int default_escape_delay = 450;	/* overridden by $ESCDELAY */
 
 static
 int dont_suppress_keymap = 1;
 
+/* The flag queue_mode_mapped indicates when the byte input queue 
+   has been stuffed with keyboard mapped characters, always being 
+   encoded in UTF-8, while otherwise the queue contents are in 
+   terminal native encoding (which may also be UTF-8, or CJK, or Latin-1)
+ */
+static
+FLAG queue_mode_mapped = False;	/* queue contains mapped character(s) */
+
 
 /*======================================================================*\
 |*		Handling function key and keyboard mapping tables	*|
@@ -33,6 +51,14 @@ int dont_suppress_keymap = 1;
 
 #ifndef msdos
 
+#define in_status_line	(current_cursor_y == YMAX)
+
+static
+void
+dummyfunc ()
+{
+}
+
 extern struct fkeyentry {
 	char * fk;
 	voidfunc fp;
@@ -62,25 +88,15 @@ set_fkeymap (term)
   }
 }
 
-#endif
-
-
-#define stringtables
-#ifdef stringtables
-#define keymaptabletype char *
 #else
-#define keymaptabletype struct keymap *
-struct keymap {
-	char * fk;
-	char * fp;
-};
-#endif
 
-extern struct {
-	keymaptabletype table;
-	char * shortcut;
-} keymappingtable [];
-extern int keymappingtable_len;
+void
+set_fkeymap (term)
+  char * term;
+{
+}
+
+#endif
 
 
 char * keyboard_mapping = "--";
@@ -122,7 +138,7 @@ toggleKEYMAP ()
 	}
 	displayflags ();
   } else {
-	error ("Keyboard mapping not active", NIL_PTR);
+	error ("Keyboard mapping not active");
   }
 }
 
@@ -174,7 +190,12 @@ setKEYMAP (script)
 	script ++;
 	selectKEYMAP (script);
 	selectKEYMAP (NIL_PTR);
-  } else {
+  } else if (script != NIL_PTR) {
+	char * secondary = strchr (script, '-');
+	if (secondary != NIL_PTR) {
+		secondary ++;
+		selectKEYMAP (secondary);
+	}
 	selectKEYMAP (script);
   }
 }
@@ -200,10 +221,10 @@ setupKEYMAP ()
 		hop_flag = 0;
 		selectKEYMAP (next_keyboard_mapping);
 	} else {
-		handleKeymapmenu (0);
+		handleKeymapmenu ();
 	}
   } else {
-	error ("Keyboard mapping not active", NIL_PTR);
+	error ("Keyboard mapping not active");
   }
 }
 
@@ -232,13 +253,14 @@ get_string_nokeymap (prompt, inbuf, statfl, term_input)
 |*		Digits input						*|
 \*======================================================================*/
 
-/*
- * get_digits () reads in a number. In contrast to get_number, it does no
- * echoing, no messaging, and it does not require any digits at all.
- * The last character typed in is returned.
- * The resulting number is put into the integer the arguments points to.
- * It is for use within I/O routines (e.g. reading mouse escape sequences 
- * or cursor position report) and thus calls _readchar ().
+/**
+   get_digits () reads in a number.
+   In contrast to get_number, it does no echoing, no messaging, reads 
+   only decimal, and accepts zero length input.
+   The last character typed in is returned.
+   The resulting number is put into the integer the arguments points to.
+   It is for use within I/O routines (e.g. reading mouse escape sequences 
+   or cursor position report) and thus calls _readchar ().
  */
 int
 get_digits (result)
@@ -264,7 +286,7 @@ get_digits (result)
 	}
   }
 
-  if (quit == True) {
+  if (quit) {
 	return quit_char;
   }
   return index;
@@ -337,9 +359,7 @@ int
 readchar ()
 {
   register character c;
-#ifdef pc_charset
   int i;
-#endif
 
 #ifdef msdos
 #define winchhere
@@ -348,7 +368,7 @@ readchar ()
 #ifdef winchhere
   FLAG waiting;
 
-  if (winchg == True && waitingforinput == False) {
+  if (winchg && waitingforinput == False) {
 	/* In the Unix version, this is now done in __readchar () */
 	RDwin ();
   }
@@ -373,24 +393,24 @@ readchar ()
 #endif
 
   /* the modification
-       if (quit == True) c = quit_char;
+	if (quit) {
+		c = quit_char;
+	}
      (now in __readchar) must not be placed after resetting the flag
-       waitingforinput = False;
-     .
+	waitingforinput = False;
      Otherwise a QUIT signal coming in just between these two would
-     discard the last valid character just taken up. */
+     discard the last valid character just taken up.
+  */
 
   timestamp_readchar ();
 
-#ifdef pc_charset
-  if (utf8_text == True && utf8_input == False && c >= 0x80) {
+  if (pc_term && utf8_text && utf8_input == False && c >= 0x80) {
 	for (i = 0; i < 128; i ++) {
 		if (c == (character) pc_charmap [i]) {
 			return (character) (i + 128);
 		}
 	}
   }
-#endif
 
   return c;
 }
@@ -424,40 +444,20 @@ readcharacter_allbuttons ()
   return readcharacter ();
 }
 
-/*
-   readcharacter () reads one Unicode character, suppressing keyboard mapping
- */
-unsigned long
-readcharacter ()
-{
-  return readcharacter_mapping (False);
-}
-
 /*
    readcharacter_mapping () reads one Unicode character
    either with or without keyboard mapping
  */
+static
 unsigned long
 readcharacter_mapping (map_keyboard)
   FLAG map_keyboard;
-{
-  return readcharacter_mapping_valid (map_keyboard, False);
-}
-
-/*
-   readcharacter_mapping () reads one Unicode character
-   either with or without keyboard mapping
- */
-unsigned long
-readcharacter_mapping_valid (map_keyboard, valid)
-  FLAG map_keyboard;
-  FLAG valid;
 {
   int utfcount = 1;
   unsigned long unichar;
   int byte2;
 
-  if (map_keyboard == True) {
+  if (map_keyboard) {
 	unichar = readchar ();
   } else {
 	unichar = readchar_nokeymap ();
@@ -469,33 +469,10 @@ readcharacter_mapping_valid (map_keyboard, valid)
 	}
   }
 
-  if ((unichar & 0x80) == 0x00) {
-	return unichar;
-  } else if (cjk_term == True && multichar (unichar)) {
-	if (cjk_encoding == 'C' && unichar == 0x8E) {
-		return unichar << 24
-			| (readchar_nokeymap () << 16)
-			| (readchar_nokeymap () << 8)
-			| readchar_nokeymap ();
-	} else if (cjk_encoding == 'J' && unichar == 0x8F) {
-		return unichar << 16
-			| (readchar_nokeymap () << 8)
-			| readchar_nokeymap ();
-	} else {
-		byte2 = readchar_nokeymap ();
-		if (cjk_encoding == 'G' && '0' <= byte2 && byte2 <= '9') {
-			return unichar << 24
-				| (byte2 << 16)
-				| (readchar_nokeymap () << 8)
-				| readchar_nokeymap ();
-		} else {
-			return unichar << 8 | byte2;
-		}
-	}
-  } else if (utf8_input == False) {
-	return unichar;
-  } else {
-	if ((unichar & 0xE0) == 0xC0) {
+  if (utf8_input || queue_mode_mapped) {
+	if ((unichar & 0x80) == 0x00) {
+		utfcount = 1;
+	} else if ((unichar & 0xE0) == 0xC0) {
 		utfcount = 2;
 		unichar = unichar & 0x1F;
 	} else if ((unichar & 0xF0) == 0xE0) {
@@ -517,18 +494,57 @@ readcharacter_mapping_valid (map_keyboard, valid)
 		unichar = (unichar << 6) | (readchar_nokeymap () & 0x3F);
 		utfcount --;
 	}
-	if (cjk_text == True || visciimode ()) {
-		unichar = cjk (unichar, True);
-		if (unichar == quit_char && valid == True) {
-			return readcharacter_mapping_valid (map_keyboard, valid);
+	if (cjk_text || mapped_text) {
+		return cjk (unichar);
+	} else {
+		return unichar;
+	}
+  } else if (cjk_term && multichar (unichar)) {
+	if (cjk_encoding == 'C' && unichar == 0x8E) {
+		return unichar << 24
+			| (readchar_nokeymap () << 16)
+			| (readchar_nokeymap () << 8)
+			| readchar_nokeymap ();
+	} else if (cjk_encoding == 'J' && unichar == 0x8F) {
+		return unichar << 16
+			| (readchar_nokeymap () << 8)
+			| readchar_nokeymap ();
+	} else {
+		byte2 = readchar_nokeymap ();
+		if (cjk_encoding == 'G' && '0' <= byte2 && byte2 <= '9') {
+			return unichar << 24
+				| (byte2 << 16)
+				| (readchar_nokeymap () << 8)
+				| readchar_nokeymap ();
 		} else {
-			return unichar;
+			return unichar << 8 | byte2;
 		}
 	}
+  } else {
 	return unichar;
   }
 }
 
+/*
+   readcharacter () reads one Unicode character
+   keyboard mapping is suppressed
+ */
+unsigned long
+readcharacter ()
+{
+  return readcharacter_mapping (False);
+}
+
+/*
+   readcharacter_mapped () reads one Unicode character
+   keyboard mapping is involved
+ */
+unsigned long
+readcharacter_mapped ()
+{
+  return readcharacter_mapping (True);
+}
+
 /*-------------------------------------------------------------------------*/
 
 /*
@@ -542,9 +558,7 @@ DIRECTxtermgetxy (code)
   character button;
   int xpos;
   int ypos;
-#ifdef mouse_hilite_tracking
-  char track [22];
-#endif
+  char track [22];	/* for mouse hilite tracking */
 
   mouse_lastbutton = mouse_button;
   if (mouse_button == leftbutton || mouse_button == rightbutton || mouse_button == middlebutton) {
@@ -603,11 +617,10 @@ DIRECTxtermgetxy (code)
   }
   mouse_ypos = ypos - 33 - MENU;
 
-#ifdef mouse_hilite_tracking
-  if (mouse_button == leftbutton) {
+  if (mouse_hilite_tracking && mouse_button == leftbutton) {
 	/* handle hilite tracking */
 	if (mouse_ypos < 0 || mouse_ypos > last_y 
-	    || (mouse_xpos == XMAX && disp_scrollbar == True) 
+	    || (mouse_xpos == XMAX && disp_scrollbar) 
 	    || is_menu_open ())
 	{
 		/* abort hilite tracking */
@@ -621,7 +634,6 @@ DIRECTxtermgetxy (code)
 		flush ();
 	}
   }
-#endif
 }
 
 static
@@ -723,7 +735,11 @@ CURSgetxy ()
 	button = 3;
   }
 
+#ifdef BUTTON_SHIFT
   mouse_shift = (BUTTON_STATUS (button) & BUTTON_MODIFIER_MASK) == BUTTON_SHIFT;
+#else
+  mouse_shift = 0;
+#endif
 
   if ((BUTTON_STATUS (button) & BUTTON_ACTION_MASK) == BUTTON_PRESSED) {
 	if (button == 1) {
@@ -749,10 +765,16 @@ static
 void
 CURSgetxy ()
 {
+#ifdef NCURSES_VERSION
   MEVENT mouse_info;
 # define mouse_state mouse_info.bstate
 # define mouse_x mouse_info.x
 # define mouse_y mouse_info.y
+#else
+  unsigned long mouse_state;
+# define mouse_x Mouse_status.x
+# define mouse_y Mouse_status.y
+#endif
 
   mouse_lastbutton = mouse_button;
   if (mouse_button == leftbutton || mouse_button == rightbutton || mouse_button == middlebutton) {
@@ -761,9 +783,18 @@ CURSgetxy ()
   mouse_lastxpos = mouse_xpos;
   mouse_lastypos = mouse_ypos;
 
+#ifdef NCURSES_VERSION
   getmouse (& mouse_info);
+#else
+  mouse_state = getmouse ();
+#endif
 
+#ifdef BUTTON_SHIFT
   mouse_shift = mouse_state & BUTTON_SHIFT;
+#else
+  mouse_shift = 0;
+#endif
+
   if (mouse_state & BUTTON1_CLICKED) {
 	mouse_button = leftbutton;
   } else if (mouse_state & BUTTON2_CLICKED) {
@@ -799,7 +830,7 @@ CURSgetxy ()
  * sequence can be detected or rejected.
  * If the queue is not empty, queue [0] contains the character next 
  * to be delivered by _readchar () (it's not a ring buffer).
- * The queue contents is always terminated by a '\0', so queue can also 
+ * The queue contents are always terminated by a '\0', so queue can also 
  * be taken as a character string.
  */
 static character queue [MAXCODELEN + 1], * endp = queue;
@@ -889,6 +920,12 @@ q_get ()
 	endp --;
 	endcp --;
   }
+
+#define dont_debug_queue_delivery
+#ifdef debug_queue_delivery
+  printf ("%02X\n", c);
+#endif
+
   return c;
 }
 
@@ -928,10 +965,30 @@ findkeyin (str, fkeymap, matchmode, found)
 
   i = lastmatch;
   do {
-	if (strncmp (str, fkeymap [i].fk, strlen (str)) == 0) {
+	char * strpoi = str;
+	char * mappoi = fkeymap [i].fk;
+	do {	/* compare str with map entry, ignore ";[0-9]" after ESC [ */
+		if (* strpoi == ';' && str [1] == '[') {
+			strpoi ++;
+			if (* strpoi >= '0' && * strpoi <= '9') {
+				strpoi ++;
+			}
+		}
+		if (* strpoi != * mappoi) {
+			break;
+		}
+		if (* strpoi) {
+			strpoi ++;
+			mappoi ++;
+		}
+	} while (* strpoi);
+
+/*	if (strncmp (str, fkeymap [i].fk, strlen (str)) == 0) {*/
+	if (* strpoi == '\0') {
 		/* str is prefix of current entry */
 		lastmatch = i;
-		if (strlen (str) == strlen (fkeymap [i].fk)) {
+/*		if (strlen (str) == strlen (fkeymap [i].fk)) {*/
+		if (* mappoi == '\0') {
 			/* str is equal to current entry */
 			* found = i;
 			keyproc = fkeymap [i].fp;
@@ -991,7 +1048,7 @@ mapkb (str, kbmap, matchmode, found, mapped)
   char * * found;
   char * * mapped;
 {
-#ifdef stringtables
+#ifdef keymapping_stringtables
 
   int lastmatch = 0;	/* last index with string matching prefix */
   register int i;
@@ -1105,9 +1162,9 @@ findkey (str)
   char * str;
 {
   int dummy;
-  int res = findkeyin (str, fkeymap, 0, & dummy);
+  int res = findkeyin (str, fkeymap2, 0, & dummy);
   if (res == -2) {
-	return findkeyin (str, fkeymap2, 0, & dummy);
+	return findkeyin (str, fkeymap, 0, & dummy);
   } else {
 	return res;
   }
@@ -1119,6 +1176,8 @@ findkey (str)
 #ifdef CURSES
 chtype key;
 
+extern int curs_readchar _((void));
+
 static
 void
 showfkey ()
@@ -1184,7 +1243,7 @@ struct {
 	{KEY_F(10), FILEMENU},
 	{KEY_F(20), QUICKMENU},
 	{KEY_F(11), LOWCAP},
-	{KEY_F(12), UML},
+	{KEY_F(12), (voidfunc) UML},
 #ifdef __PDCURSES__
 	{PADSLASH, I},
 	{PADSTAR, I},
@@ -1275,10 +1334,11 @@ putback_rest (rest)
   char old_rest_queue [MAXCODELEN + 1];
 
 	if (! streq (rest, " ")) {
-		if (have_rest_queue == True) {
+		if (have_rest_queue) {
 			/**
-			   overdraft characters to be mapped are put back into 
-			   rest_queue to be considered for subsequent mapping
+			   overdraft characters to be mapped are 
+			   put back into rest_queue 
+			   to be considered for subsequent mapping
 				-> Hiragana nihongo
 				-> Hiragana nyi
 				-> test abcde
@@ -1295,13 +1355,16 @@ putback_rest (rest)
 
 #endif
 
-character
+/* read 1 byte from queue; for curses, however, function 
+   key indications must be passed on which are longer than 8 bits
+ */
+int
 read1byte ()
 {
 #ifndef msdos
   character ch;
   character * cr;
-  if (have_rest_queue == True) {
+  if (have_rest_queue) {
 	ch = * rest_queue;
 	cr = rest_queue;
 	while (* (cr + 1) != '\0') {
@@ -1331,7 +1394,7 @@ int
 bytereadyafter (msec)
   int msec;
 {
-  return (have_rest_queue == True) || inputreadyafter (input_fd, msec);
+  return have_rest_queue || inputreadyafter (input_fd, msec);
 }
 
 /*
@@ -1355,13 +1418,73 @@ char_ready_within (msec)
 
 #endif
 
+
+#ifndef msdos
+#define radical_stroke
+#endif
+
+
+#ifdef radical_stroke
+static menuitemtype radicalsmenu [] =
+{
+	/* 1 stroke radicals */
+	{"①-1:一 2:丨 3:丶 4:丿乀乁 5:乙乚乛 6:亅", dummyfunc, ""},
+	/* 2 strokes radicals */
+	{"②-7:二 8:亠 9:人亻 10:儿 11:入 12:八 13:冂 14:冖", dummyfunc, ""},
+	{"②-15:冫 16:几 17:凵 18:刀刁刂 19:力 20:勹 21:匕", dummyfunc, ""},
+	{"②-22:匚 23:匸 24:十 25:卜 26:卩 27:厂 28:厶 29:又", dummyfunc, ""},
+	/* 3 strokes radicals */
+	{"③-30:口 31:囗 32:土 33:士 34:夂 35:夊 36:夕 37:大夨 38:女 39:子孑孒孓", dummyfunc, ""},
+	{"③-40:宀 41:寸 42:小 43:尢尣 44:尸 45:屮 46:山 47:巛巜川 48:工 49:己已巳", dummyfunc, ""},
+	{"③-50:巾 51:干 52:乡幺 53:广 54:廴 55:廾 56:弋 57:弓 58:彐彑 59:彡 60:彳", dummyfunc, ""},
+	/* 4 strokes radicals */
+	{"④-61:心忄 62:戈 63:戶户戸 64:手扌才 65:支 66:攴攵 67:文 68:斗", dummyfunc, ""},
+	{"④-69:斤 70:方 71:无 72:日 73:曰 74:月 75:木朩 76:欠 77:止 78:歹 79:殳", dummyfunc, ""},
+	{"④-80:毋毌 81:比 82:毛 83:氏 84:气 85:水氵 86:火灬 87:爪爫 88:父 89:爻", dummyfunc, ""},
+	{"④-90:丬爿 91:片 92:牙㸦 93:牛牜 94:犬犭", dummyfunc, ""},
+	/* 5 strokes radicals */
+	{"⑤-95:玄 96:玉王 97:瓜 98:瓦 99:甘 100:生 101:用甩 102:田由甲申甴电", dummyfunc, ""},
+	{"⑤-103:疋 104:疒 105:癶 106:白 107:皮 108:皿 109:目", dummyfunc, ""},
+	{"⑤-110:矛 111:矢 112:石 113:示礻 114:禸 115:禾 116:穴 117:立", dummyfunc, ""},
+	/* 6 strokes radicals */
+	{"⑥-118:竹 119:米 120:糸糹纟 121:缶 122:网罒罓䍏 123:羊 124:羽 125:老耂考", dummyfunc, ""},
+	{"⑥-126:而 127:耒 128:耳 129:聿肀 130:肉 131:臣 132:自", dummyfunc, ""},
+	{"⑥-133:至 134:臼 135:舌 136:舛 137:舟 138:艮 139:色 140:艸艹䒑", dummyfunc, ""},
+	{"⑥-141:虍 142:虫 143:血 144:行 145:衣衤 146:襾西覀", dummyfunc, ""},
+	/* 7 strokes radicals */
+	{"⑦-147:見见 148:角 149:言訁讠 150:谷 151:豆 152:豕 153:豸", dummyfunc, ""},
+	{"⑦-154:貝贝 155:赤 156:走赱 157:足 158:身 159:車车 160:辛", dummyfunc, ""},
+	{"⑦-161:辰 162:辵辶 163:邑 164:酉 165:釆 166:里", dummyfunc, ""},
+	/* 8 strokes radicals */
+	{"⑧-167:金釒钅 168:長镸长 169:門门 170:阜阝", dummyfunc, ""},
+	{"⑧-171:隶 172:隹 173:雨 174:靑青 175:非", dummyfunc, ""},
+	/* 9 strokes radicals */
+	{"⑨-176:面靣 177:革 178:韋韦 179:韭 180:音 181:頁页", dummyfunc, ""},
+	{"⑨-182:風风 183:飛飞 184:食飠饣 185:首 186:香", dummyfunc, ""},
+	/* 10 strokes radicals */
+	{"⑩-187:馬马 188:骨 189:高髙 190:髟 191:鬥 192:鬯 193:鬲 194:鬼", dummyfunc, ""},
+	/* 11 strokes radicals */
+	{"⑪-195:魚鱼 196:鳥鸟 197:鹵 198:鹿 199:麥麦 200:麻", dummyfunc, ""},
+	/* 12 strokes radicals */
+	{"⑫-201:黃黄 202:黍 203:黑黒 204:黹", dummyfunc, ""},
+	/* 13 strokes radicals */
+	{"⑬-205:黽黾 206:鼎 207:鼓鼔 208:鼠鼡", dummyfunc, ""},
+	/* 14 strokes radicals */
+	{"⑭-209:鼻 210:齊齐", dummyfunc, ""},
+	/* 15 strokes radicals */
+	{"⑮-211:齒齿", dummyfunc, ""},
+	/* 16 strokes radicals */
+	{"⑯-212:龍龙 213:龜龟", dummyfunc, ""},
+	/* 17 strokes radical */
+	{"⑰-214:龠", dummyfunc, ""},
+};
+#endif
+
+
 int
 _readchar ()
 {
   register character ch;
-#ifndef msdos
-  char second_esc_char;
-#endif
 #ifdef dosmouse
   int ich;
 #endif
@@ -1373,21 +1496,108 @@ _readchar ()
 #endif
   char * choice;
   int choice_index;
-#ifndef msdos
-  unsigned long unichar;
-  int utflen;
-  character cjkbytes [5];
-#endif
+
+
+  if (escape_delay == 0 || keymap_delay == 0) {
+	char * env = getenv ("ESCDELAY");
+	if (env) {
+		(void) scan_int (env, & escape_delay);
+	}
+	if (escape_delay == 0) {
+		escape_delay = default_escape_delay;
+	}
+	env = getenv ("MAPDELAY");
+	if (env) {
+		(void) scan_int (env, & keymap_delay);
+	}
+	if (keymap_delay == 0) {
+		keymap_delay = default_keymap_delay;
+	}
+  }
 
 
 #ifndef msdos
   if (q_len () > 0) {
 	return q_get ();
   }
+  queue_mode_mapped = False;
 #endif
 
   keyshift = '0';
 
+
+#ifdef radical_stroke
+  /* plug-in specific input methods */
+  if (allow_keymap && dont_suppress_keymap
+	&& streq (keyboard_mapping, "rs")
+  ) {
+	int choice_col = x;
+	int choice_line = y + 1;
+	char radicalnumber [4];
+
+	choice_index = -1;
+
+	/* adjust menu position if too far down */
+	if (choice_line + arrlen (radicalsmenu) + 2 > YMAX) {
+		choice_line = y - arrlen (radicalsmenu) - 2;
+		if (choice_line < 0) {
+			choice_line = 0;
+		}
+	}
+
+	/* adjust menu position if on status line */
+	if (in_status_line) {
+		choice_line = YMAX - arrlen (radicalsmenu) - 2;
+		if (choice_line < 0) {
+			choice_line = 0;
+		}
+		choice_col = lpos;
+	}
+
+	/* try to get radical, then character, from user */
+	while (choice_index < 0) {
+		int radical = popup_menu (radicalsmenu, arrlen (radicalsmenu), 
+					choice_col, choice_line, 
+					"Select radical", False, "1234567890x");
+		if (radical < 0) {
+			/* revert to previous mapping (or none) */
+			toggleKEYMAP ();
+			set_cursor_xy ();
+			flush ();
+			choice_index = 0;	/* continue */
+		} else {
+			int radical_row = radical / 11;
+			int radical_col = radical % 11;
+			mapped = radicalsmenu [radical_row].itemname;
+			while (* mapped != '\0' &&
+			       (* mapped < '0' || * mapped > '9'
+			        || radical_col > 0))
+			{
+				if (* mapped == ':') {
+					radical_col --;
+				}
+				mapped ++;
+			}
+			found = radicalnumber;
+			while (* mapped >= '0' && * mapped <= '9') {
+				* found ++ = * mapped ++;
+			}
+			* found = '\0';
+			res = mapkb (radicalnumber, keyboard_map, 2, & found, & mapped);
+			if (res >= 0) {
+				/* first goto in my life and then 
+				   such an exceptionally ugly one -
+				   but it works well just this way */
+				goto handle_mapped;
+			} else {
+				ring_bell ();
+			}
+		}
+	}
+  }
+#endif
+
+
 #ifdef CURSES
 
   key = read1byte ();
@@ -1454,7 +1664,7 @@ _readchar ()
 # else
 	ch = read1byte ();
 # endif
-	keyproc = pc_key_map [ch];
+	keyproc = pc_fkey_map [ch];
 # ifdef DEBUG
 	if ((voidfunc) keyproc == (voidfunc) I) {
 		return ch;
@@ -1475,34 +1685,86 @@ _readchar ()
 #ifndef msdos
 
     else if (ch == '\033') {
+	char second_esc_char = '\0';
+
 	q_put (ch);
-	second_esc_char = ch;
+
 	while  ((res = findkey (ctrl_queue)) == -1 /* prefix of table entry */
 		&& q_notfull ()
 		&& bytereadyafter (escape_delay)
 	       )
 	{
+		char prev_ch = ch;
 		ch = read1byte ();
 
 		/* handle generic shift state of ANSI sequences: */
 		if (ch == ';' && second_esc_char == '[') {
-			keyshift = read1byte () - 1;
+			/* ESC [1;2A -> ESC [1A + shift etc */
+			q_put (';');
+			ch = read1byte ();
+			keyshift = ch - 1;
+			q_put (ch);
 			ch = read1byte ();
 		} else if (ch >= '0' && ch <= '9' && second_esc_char == 'O') {
+			/* ESC O2A -> ESC OA + shift etc */
 			keyshift = ch - 1;
 			ch = read1byte ();
+		} else if (prev_ch >= '0' && prev_ch <= '9') {
+			if (ch == '^') {
+				/* ESC [5^ -> ESC [5~ + control (rxvt) */
+				keyshift = ctrl_mask;
+				ch = '~';
+			} else if (ch == '$') {
+				/* ESC [5$ -> ESC [5~ + shift (rxvt) */
+				keyshift = shift_mask;
+				ch = '~';
+			} else if (ch == '@') {
+				/* ESC [5@ -> ESC [5~ + control+shift (rxvt) */
+				keyshift = ctrlshift_mask;
+				ch = '~';
+			}
+		} else if (ch >= 'a' && ch <= 'd') {
+			if (prev_ch == 'O') {
+				/* ESC Oa -> ESC OA + control (rxvt) */
+				keyshift = ctrl_mask;
+				ch -= 'a' - 'A';
+			} else if (prev_ch == '[') {
+				/* ESC [a -> ESC [A + shift (rxvt) */
+				keyshift = shift_mask;
+				ch -= 'a' - 'A';
+			}
 		}
 
 		q_put (ch);
-		if (second_esc_char == '\033') {
+		if (second_esc_char == '\0') {
 			second_esc_char = ch;
 		}
 	}
 
-	if (quit == True) {
-		return '\0';
+	if (quit) {
+		keyproc = I;
+		return FUNcmd;
 	} else if (res < 0) { /* key pattern not detected in fkeymap table */
-		return q_get () /* just deliver the typed characters */;
+		if (second_esc_char == '[') {
+			/* generic scanning for unknown ANSI sequence */
+			(void) q_get (); /* swallow ESC */
+			(void) q_get (); /* swallow [ */
+			ansi_params = 0;
+			do {
+				int i;
+				ansi_fini = get_digits (& i);
+				if (ansi_params < max_ansi_params) {
+					ansi_param [ansi_params] = i;
+					ansi_params ++;
+				}
+			} while (ansi_fini == ';');
+
+			keyproc = ANSIseq;
+			return FUNcmd;
+		} else {
+			/* just deliver the typed characters */
+			return q_get ();
+		}
 	} else {
 		q_clear ();
 
@@ -1568,6 +1830,7 @@ _readchar ()
 	printf ("res %d\n", res);
 #endif
 	if (res >= 0) { /* key pattern detected in key mapping table */
+handle_mapped:
 		q_clear ();
 
 		/* check if mapping defines multiple choices */
@@ -1603,26 +1866,11 @@ _readchar ()
 		}
 
 		if (mapped != NIL_PTR) {
-		    if (cjk_text == True && utf8_input == False) {
-			utf8_info (mapped, & utflen, & unichar);
-#ifdef debug_kbmap
-	printf (" mapped %04X -> %04X\n", unichar, cjk (unichar, False));
-#endif
-			unichar = cjk (unichar, True);
-			if (unichar != quit_char) {
-				(void) cjkencode (unichar, cjkbytes);
-				mapped = (char *) cjkbytes;
-				while (* mapped != '\0') {
-					q_put (* mapped);
-					mapped ++;
-				}
-			}
-		    } else {
+			queue_mode_mapped = True;
 			while (* mapped != '\0' && * mapped != ' ') {
 				q_put (* mapped);
 				mapped ++;
 			}
-		    }
 		}
 #ifdef debug_kbmap
 	printf (" queued %s\n", ctrl_queue);
@@ -1647,12 +1895,6 @@ _readchar ()
 
 #ifndef msdos
 
-static
-void
-dummyfunc ()
-{
-}
-
 #define dont_debug_choose_char
 
 static
@@ -1674,7 +1916,9 @@ choose_char (choices)
 
   int selected;
 
-  int in_status_line = current_cursor_y == YMAX;
+#ifdef cut_picklist_to_screen
+  FLAG reduced_menu = False;
+#endif
 
 	/* count choices */
 	do {
@@ -1691,7 +1935,7 @@ choose_char (choices)
 
 	lastrow = (choice_count - 1) / 10;
 	choicelines = alloc (strlen (choices) + 1 + choice_count * 2);
-	choicemenu = (menuitemtype *) alloc ((lastrow + 1) * sizeof (menuitemtype));
+	choicemenu = alloc ((lastrow + 1) * sizeof (menuitemtype));
 	if (choicelines == NIL_PTR || choicemenu == (menuitemtype *) NIL_PTR) {
 		if (choicelines != NIL_PTR) {
 			free_space (choicelines);
@@ -1762,12 +2006,26 @@ choose_char (choices)
 		thischoiceline - choicelines);
 #endif
 
+#ifdef cut_picklist_to_screen
+	if (lastrow >= YMAX - 2) {
+		lastrow = YMAX - 3;
+		reduced_menu = True;
+		if (! in_status_line) {
+			status_msg ("Warning: full menu too large to display");
+		}
+	}
+#endif
 	selected = popup_menu (choicemenu, lastrow + 1, 
 				choice_col, choice_line, 
-				"Choose character", "1234567890");
+				"Choose character", False, "1234567890");
 	if (in_status_line) {
 		redraw_prompt ();
+#ifdef cut_picklist_to_screen
+	} else if (reduced_menu) {
+		clear_status ();
+#endif
 	}
+
 #ifdef debug_choose_char
 	printf ("selected %d\n", selected);
 #endif
