@@ -7,243 +7,6 @@
 #include "io.h"
 
 
-/*======================================================================*\
-|*			8 bit character constant stuff			*|
-\*======================================================================*/
-
-char TABdefault = '';		/* default TAB indicator */
-char RETdefault = '';		/* indicates line end */
-char DOSRETdefault = '';	/* indicates DOS line end */
-char PARAdefault = '';		/* indicates end of paragraph */
-char SHIFT_MARK = '';		/* indicates that line continues */
-char UNI_MARK = '';		/* to be shown in place of Unicode char */
-
-int
-idfchar (cpoi)
-  character * cpoi;
-{
-  int ret;
-  unsigned long unichar = charvalue (cpoi);
-
-  ret =
-	 ('0' <= unichar && unichar <= '9') ||
-	 ('A' <= unichar && unichar <= 'Z') ||
-	 ('a' <= unichar && unichar <= 'z') ||
-	 (utf8_text == True ?
-	   (unichar >= 0x00C0 && unichar < 0x2000) ||
-	   (unichar >= 0x2800 && unichar < 0x3000) ||
-	   (unichar >= 0x3040 && unichar < 0xFE00) ||
-	   (unichar >= 0xFE70 && unichar < 0xFEFF) ||
-	   (unichar >= 0xFE70 && unichar < 0xFEFF) ||
-	   (unichar >= 0xFF10 && unichar < 0xFF1A) ||
-	   (unichar >= 0xFF21 && unichar < 0xFF3B) ||
-	   (unichar >= 0xFF41 && unichar < 0xFF5B) ||
-	   (unichar >= 0xFF66 && unichar < 0xFFE0) ||
-	   (unichar >= 0x10000 && unichar < 0xE0000) ||
-	   unichar >= 0xF0000
-	  :
-#ifdef pc_charset
-	     ((character) '\200' <= unichar && unichar <= (character) '\232') ||
-	     ((character) '\240' <= unichar && unichar <= (character) '\245') ||
-	     ((character) '\340' <= unichar && unichar <= (character) '\353')
-#else
-	     ((character) '\300' <= unichar && unichar != (character) '' && unichar != (character) '') ||
-	     (unichar == (character) '')
-#endif
-	 ) ||
-	 (unichar == '_');
-  return ret;
-}
-
-
-/*======================================================================*\
-|*			MSDOS Video mode switching table		*|
-\*======================================================================*/
-
-struct {
-	int mode, cols, lins;
-} modetab [] =
-{ /* list of available modes, unsorted, terminated by a '-1' mode */
-  /* the entries contain:
-	a video mode number (restriction to text modes preferable),
-	the numbers of columns and of lines which can be displayed 
-		in that mode if a font of height 16 is used */
-  {42, 100, 40},
-  {38, 80, 60},
-  {36, 132, 28},
-  {35, 132, 25},
-  {34, 132, 44},
-  { 3, 80, 25},
-  { 1, 40, 25},
-  {-1, 0, 0}
-};
-
-
-/*======================================================================*\
-|*			Character composition				*|
-\*======================================================================*/
-
-unsigned long
-diaeresis (ch)
-  unsigned long ch;
-{
-  if (keyshift & ctrl_mask) {
-	keyshift = '0';
-	return angstrom (ch);
-  } else if (keyshift & shift_mask) {
-	keyshift = '0';
-	return tilde (ch);
-  } else {
-	return compose ('"', ch);
-  }
-}
-
-unsigned long
-acute (ch)
-  unsigned long ch;
-{
-  if (keyshift & ctrl_mask) {
-	keyshift = '0';
-	return circumflex (ch);
-  } else if (keyshift & shift_mask) {
-	keyshift = '0';
-	return grave (ch);
-  } else {
-	return compose ('\'', ch);
-  }
-}
-
-unsigned long
-tilde (ch)
-  unsigned long ch;
-{
-  return compose ('~', ch);
-}
-
-unsigned long
-grave (ch)
-  unsigned long ch;
-{
-  return compose ('`', ch);
-}
-
-unsigned long
-angstrom (ch)
-  unsigned long ch;
-{
-  return compose (UNI_ring, ch);
-}
-
-unsigned long
-circumflex (ch)
-  unsigned long ch;
-{
-  return compose ('^', ch);
-}
-
-
-struct {
-	char * mnemo;
-	int uni;
-} unimnemo3 [] =
-{
-#include "mnemos.h"
-	{NIL_PTR}
-};
-
-unsigned long
-compose_mnemo (mnemo)
-  char * mnemo;
-{
-  int i = 0;
-  char * mnemo2 = & mnemo [1];
-  unsigned long ret;
-
-  while (unimnemo3 [i].mnemo != NIL_PTR) {
-	if (strcmp (unimnemo3 [i].mnemo, mnemo) == 0
-	    || (mnemo [0] == '&' && strcmp (unimnemo3 [i].mnemo, mnemo2) == 0)
-	   )
-	{
-		ret = unimnemo3 [i].uni;
-		if (cjk_text == True || isviscii (ret)) {
-			return cjk (ret, True);
-		} else {
-			return ret;
-		}
-	}
-	i ++;
-  }
-
-  return 0;
-}
-
-/**
-   Normalize accent shortcuts to usually available keys as used in tables.
- */
-static
-void
-subst_acc (c1p, c2p, acc, sub)
-  unsigned long * c1p;
-  unsigned long * c2p;
-{
-  if (* c1p == acc) {
-	* c1p = sub;
-  }
-  if (* c2p == acc) {
-	* c2p = sub;
-  }
-}
-
-unsigned long
-compose (c1, c2)
-  unsigned long c1;
-  unsigned long c2;
-{
-  character cs [13];
-  character * cp = cs;
-  unsigned long ret;
-#ifdef pc_charset
-  character c;
-#endif
-
-  subst_acc (& c1, & c2, (character) '', '\'');
-  subst_acc (& c1, & c2, (character) '', '"');
-  subst_acc (& c1, & c2, (character) '', ',');
-
-  cp += uniUTF (c1, cp);
-  uniUTF (c2, cp);
-
-  ret = compose_mnemo (cs);
-
-#define try_reverse
-#ifdef try_reverse
-  if (ret == 0) {
-	cp = cs;
-	cp += uniUTF (c2, cp);
-	uniUTF (c1, cp);
-	ret = compose_mnemo (cs);
-  }
-#endif
-
-#ifdef pc_charset
-  if (ret >= 0x80 && ret < 0x100 && utf8_text == False) {
-	c = (character) pc_charmap [ret - 0x80];
-	if (c < 0x80) {
-		return 0;
-	} else {
-		return c;
-	}
-  }
-#endif
-
-  if (utf8_text == True || cjk_text == True || ret < 256) {
-	return ret;
-  } else {
-	return 0;
-  }
-}
-
-
 /*======================================================================*\
 |*			Key and function key -> command mapping		*|
 \*======================================================================*/
@@ -270,9 +33,9 @@ voidfunc key_map [256] =
   /* ^@-^G */	NULLcmd, MPW, DCC, PD, MRT, MUP, MNW, GOTO,
   /* ^H-^O */	/*BSPcmd*/ DPC, Sch, SNL, DLN, HOP, SNL, RS, LIB,
   /* ^P-^W */	PASTEstay, HOP, PU, MLF, DNW, CUT, CTRLINS, SU,
-  /* ^X-^_ */	MDN, COPY, SD, ESCAPE, CANCEL, MARK, DPW, (voidfunc) BAD,
+  /* ^X-^_ */	MDN, COPY, SD, ESCAPE, CANCEL, MARK, DPW, (voidfunc) BADch,
   /* 040-057 */	Sch, Sch, Sdoublequote, Sch, Sch, Sch, Sch, Ssinglequote,
-		 Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
+		 Sch, Sch, Sch, Sch, Sch, Sdash, Sch, Sch,
   /* 060-077 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
   /* 100-117 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
   /* 120-137 */	Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch, Sch,
@@ -292,8 +55,8 @@ voidfunc ws_key_map [32] =
 {	/* characters to functions map */
   /* ^@-^G */	NULLcmd, MPW, JUS, PD, MRT, MUP, MNW, DCC,
   /* ^H-^O */	MLF, Sch, Sch, ctrlK, RS, SNL, LIB, ctrlO,
-  /* ^P-^W */	CTRLINS, ctrlQ, PU, MLF, DNW, (voidfunc) BAD, TOGINS, SU,
-  /* ^X-^_ */	MDN, DLINE, SD, ESCAPE, CANCEL, (voidfunc) BAD, (voidfunc) BAD, (voidfunc) BAD,
+  /* ^P-^W */	CTRLINS, ctrlQ, PU, MLF, DNW, (voidfunc) BADch, TOGINS, SU,
+  /* ^X-^_ */	MDN, DLINE, SD, ESCAPE, CANCEL, (voidfunc) BADch, (voidfunc) BADch, (voidfunc) BADch,
 };
 
 voidfunc emacs_key_map [32] =
@@ -343,7 +106,7 @@ HELPA ()
 
 
 /* mapping function key codes -> functions */
-voidfunc pc_key_map [256] =
+voidfunc pc_fkey_map [256] =
 {
    /* 00-0F */ NULLcmd, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,
    /* 10 alt-Q */ I,
@@ -514,13 +277,13 @@ struct fkeyentry {
 	{"\033k",	toggleKEYMAP},
 	{"\033K",	setupKEYMAP},
 
-	/* VT100 and SUN */
+	/* VT100 and Sun */
 	{"\033[A",	MUP},	/* move cursor up	*/
 	{"\033[B",	MDN},	/* more cursor down	*/
 	{"\033[C",	MRT},	/* more cursor right	*/
 	{"\033[D",	MLF},	/* more cursor left	*/
 
-	/* SUN raw function key codes */
+	/* Sun raw function key codes */
 	{"\033[215z",	MUP},	/* move cursor up	*/
 	{"\033[221z",	MDN},	/* more cursor down	*/
 	{"\033[219z",	MRT},	/* more cursor right	*/
@@ -537,6 +300,7 @@ struct fkeyentry {
 	{"\033[[D",	INSFILE},	/* F4	*/
 	{"\033[[E",	insert_diaeresis},	/* F5	*/
 	{"\033[G",	HOP},	/* 5	*/
+	{"\033[1G",	HOP},	/* shift-5? (or some other terminal)	*/
 
 	/* Linux X / xterm */
 	{"\033OP",	HELP},	/* F1	*/
@@ -555,6 +319,7 @@ struct fkeyentry {
 	/* xterm generic keypad keys */
 	{"\033[1H",	HOMEkey},	/* Pos1	*/
 	{"\033[1F",	ENDkey},	/* End	*/
+	{"\033[1E",	HOP},	/* 5	*/
 	{"\033[1A",	MUP},	/* Up	*/
 	{"\033[1D",	MLF},	/* Left	*/
 	{"\033[1B",	MDN},	/* Down	*/
@@ -564,6 +329,19 @@ struct fkeyentry {
 	{"\033[57~",	SNL},	/* KP_Enter	*/
 	{"\033[58~",	COPY},	/* control-Ins	*/
 
+	/* mined keyboard add-ons: shifted digits etc*/
+	{"\033[00~",	shift_0},
+	{"\033[01~",	shift_1},
+	{"\033[02~",	shift_2},
+	{"\033[03~",	shift_3},
+	{"\033[04~",	shift_4},
+	{"\033[05~",	shift_5},
+	{"\033[06~",	shift_6},
+	{"\033[07~",	shift_7},
+	{"\033[08~",	shift_8},
+	{"\033[09~",	shift_9},
+	{"\033[~",	DPC},	/* ctrl-BackSpace */
+
 	/* various modified keypad keys (non-application mode) */
 	{"\033\r",	Popmark},	/* alt-Enter	*/
 	{"\033\n",	Popmark},	/* alt-Enter	*/
@@ -618,10 +396,10 @@ struct fkeyentry {
 
 	/* VT100 - some have been disabled by linux sequences */
 	/* F1...F10 see above, F11...F20 see fkeymap_vt100 below */
-	{"\033[6~",	PD},	/* Next Screen	*/
-	{"\033[5~",	PU},	/* Prev Screen	*/
-	{"\033[3~",	DELkey},	/* Remove	*/
-	{"\033[2~",	PASTE},	/* Insert Here	*/
+	{"\033[6~",	PD},	/* Next Screen, xterm PgDn	*/
+	{"\033[5~",	PU},	/* Prev Screen, xterm PgUp	*/
+	{"\033[3~",	DELkey},	/* Remove, xterm Del	*/
+	{"\033[2~",	PASTE},	/* Insert Here, xterm Ins	*/
 /*	{"\033OP",	HOP},	*//* PF1	*/
 /*	{"\033OQ",	SU},	*//* PF2	*/
 /*	{"\033OR",	SD},	*//* PF3	*/
@@ -653,7 +431,7 @@ struct fkeyentry {
 	{"\033Op",	PASTE},	/* Ins	*/
 	{"\033On",	DELkey},	/* Del	*/
 
-	/* SUN */
+	/* Sun */
 	{"\033[208z",	insert_grave},	/* R1 */
 	{"\033[209z",	insert_circumflex},	/* R2 */
 	{"\033[210z",	insert_acute},	/* R3 */
@@ -735,6 +513,11 @@ struct fkeyentry {
 	{"\033OY",	MDN},	/* dn	*/
 	{"\033O?",	PD},	/* PgDn, End, Ins, Del	*/
 
+	/* rxvt */
+	{"\033[7~",	smallHOMEkey},	/* 	*/
+	{"\033[8~",	smallENDkey},	/* 	*/
+	{"\033[Z",	MPW},		/* shift-TAB */
+
 	{NIL_PTR}
 };
 
@@ -788,24 +571,16 @@ struct fkeyentry fkeymap_hp [] = {
 	{"\0332u",	insert_grave},	/* shift-F6 */
 	{"\0332v",	I},	/* shift-F7 */
 	{"\0332w",	GR},	/* shift-F8 */
+	{"\033V",	PU},
+	{"\033U",	PD},
+
+	{"\033&",	SNL},	/* keypad Enter */
 
 	{NIL_PTR}
 };
 
 struct fkeyentry fkeymap_siemens [] = {
-	/* VT100 emulator at IBM-PC */
-	{"\033M",	PU},	/* Pg Up */
-	{"\033D",	PD},	/* Pg Dn */
-/*	{"\033[H",	smallHOMEkey},	*//* Home	, same at Iris */
-	{"\033[24H",	smallENDkey},	/* End; actually "\033[24;1H", ";1" filtered out */
-	{"\0331",	HOP},	/* Ins	*/
-		/* cursor keys and F1-4 same as VT100 */
-
 	/* Siemens 97801/97808 */
-	{"\033A",	WT},
-	{"\033;",	WTU},
-	{"\033B",	EDIT},
-	{"\033\"",	VIEW},
 	{"\033N",	insert_diaeresis},	/* F14 */
 	{"\033P",	RS},			/* F16 */
 	{"\0332",	insert_grave},		/* shift-F18 */
@@ -819,14 +594,41 @@ struct fkeyentry fkeymap_siemens [] = {
 	{"\0339",	MPW},		/* <~- */
 	{"\033:",	MNW},		/* -~> */
 	{"\0336",	HOP},		/* neben END, ber Cursorblock */
+	{"\033[@",	PASTE},		/* Insert */
 /*
-INSERT:	[@	o	[L
-DELETE:	([P)	p	[M
-|<-:	[Z
+?
+INSERT: o
+DELETE: p [M
 */
 	{"\033[S",	PU},
 	{"\033[T",	PD},
 	{"\033^",	HOP},
+
+	{"\033@",	HELP},	/* F1 */
+	{"\033A",	WT},	/* F2 */
+	{"\033B",	EDIT},	/* F3 */
+	{"\033C",	INSFILE},	/* F4 */
+	{"\033D",	PD},	/* F5 , Pg Dn ? */
+	{"\033F",	insert_acute},	/* F6 */
+	{"\033G",	SFW},	/* F7 */
+	{"\033H",	SFW},	/* F8 */
+	{"\033I",	RS},	/* F9 */
+	{"\033J",	FILEMENU},	/* F10 */
+	{"\033K",	LOWCAP},	/* F11 */
+	{"\033L",	SAVPOS},	/* F12 */
+	{"\033 ",	I},	/* shift-F1 */
+	{"\033;",	WTU},	/* shift-F2 */
+	{"\033\"",	I},	/* shift-F3 */
+	{"\033#",	WB},	/* shift-F4 */
+	{"\033$",	insert_tilde},	/* shift-F5 */
+	{"\033%",	insert_grave},	/* shift-F6 */
+	{"\033&",	I},	/* shift-F7 */
+	{"\033'",	GR},	/* shift-F8 */
+	{"\033<",	I},	/* shift-F9 */
+	{"\033=",	QUICKMENU},	/* shift-F10 */
+	{"\033*",	I},	/* shift-F11 */
+	{"\033+",	I},	/* shift-F12 */
+
 	{NIL_PTR}
 };
 
