@@ -10,17 +10,6 @@
 |*		Keyboard mapping tables					*|
 \*======================================================================*/
 
-#define stringtables
-#ifdef stringtables
-#define keymaptabletype char *
-#else
-#define keymaptabletype struct keymap *
-struct keymap {
-	char * fk;
-	char * fp;
-};
-#endif
-
 #ifdef use_keymap_tables
 #include "_keymaps.h"
 #endif
@@ -30,10 +19,7 @@ struct keymap {
 |*		Table of tables						*|
 \*======================================================================*/
 
-struct {
-	keymaptabletype table;
-	char * shortcut;
-} keymappingtable [] =
+struct keymapping keymappingtable [] =
 {
 #ifdef use_keymap_tables
 #include "keymapsk.h"
