@@ -3,9 +3,11 @@
  *  This file is placed in the public domain.
  */
 
-#include <stdlib.h>
 #include <string.h>
 
+extern char * getenv ();
+
+
 int
 is_locale_utf8 ()
 {
@@ -13,11 +15,11 @@ is_locale_utf8 ()
 
   /* Determine the current locale the same way as setlocale() does, 
      according to POSIX. */
-  locale = (char *) getenv ("LC_ALL");
+  locale = getenv ("LC_ALL");
   if (! locale || ! * locale) {
-    locale = (char *) getenv ("LC_CTYPE");
+    locale = getenv ("LC_CTYPE");
     if (! locale || ! * locale) {
-      locale = (char *) getenv ("LANG");
+      locale = getenv ("LANG");
     }
   }
   if (locale && * locale) {
