@@ -34,7 +34,7 @@
 # define __GCC__
 
 #  ifdef __strange__
-#  undef msdos	/* in this queer case, a more specific distinction */
+#  undef msdos	/* in this weird case, a more specific distinction */
 #  define unix	/* has to be made everywhere according to the 'pc' flag */
 #  endif
 # endif
@@ -63,24 +63,24 @@ extern void exit ();
 #  define FUNcmd	'\000'
 
 # ifdef __TURBOC__
-/* With Turbo-C, compile with memory model >= Compact */
-/* #define conio doesn't work with current combined positioning/output method */
+/* #define conio doesn't work with combined positioning/output method */
 extern far * getenv ();
 extern far * getcwd ();
 extern void setdisk ();
 extern void sleep ();
 # else
 #define conio
+extern char * getenv ();
+extern char * getcwd ();
 # endif
 
 #endif
 
-#ifdef conio
-#define dosmouse
-#endif
-
 #ifdef __TURBOC__
 #define conio
+#endif
+
+#ifdef conio
 #define dosmouse
 #endif
 
@@ -138,18 +138,19 @@ extern void exit ();
 extern char * getenv ();
 /*extern int printf ();*/
 
-/* already declared in <unistd.h>
+#ifdef DEFVFORK
+/* already declared in <unistd.h> */
 #ifdef __STDC__
 extern pid_t vfork ();
 #else
 #ifdef _CLASSIC_ID_TYPES
-#define is_this_a_HP_special
+/* is this a HP special? */
 extern int vfork();
 #else
 extern pid_t vfork ();
 #endif
 #endif
-*/
+#endif
 
 #ifdef __CYGWIN__
 extern int vfork();
@@ -230,21 +231,19 @@ extern int scrollbar_width;
 					x-coordinate for line; MAX_CHARS + 1
 					is not sufficient due to TABs */
 
-#define BLOCK_SIZE	1024
-
 /* Return values of functions */
 #define ERRORS		-1
 #define NO_LINE		(ERRORS - 1)	/* Must be < 0 */
 #define NUL_LINE	(ERRORS - 2)	/* Must be < 0 */
 #define SPLIT_LINE	(ERRORS - 3)	/* Must be < 0 */
-#define FINE	 	(ERRORS + 1)
-#define NO_INPUT	(ERRORS + 2)
+#define FINE		0
+#define NO_INPUT	(FINE + 1)
 
 #define STD_IN		0		/* Input file # */
 #define STD_OUT	 	1		/* Terminal output file # */
 #define STD_ERR		2
 
-#define REPORT	1		/* Report change of lines on # lines */
+#define REPORT_CHANGED_LINES	1	/* Report change of lines on # lines */
 
 #ifndef pc
 #ifndef O_BINARY
@@ -337,45 +336,18 @@ typedef struct Line LINE;
 #define DUMMY		0x80
 #define DUMMY_MASK	0x7F
 
-/* Expression definitions */
-#define NO_MATCH	0
-#define MATCH		1
-#define REG_ERROR	2
-
-#define BEGIN_LINE	(2 * REG_ERROR)
-#define END_LINE	(2 * BEGIN_LINE)
-
 /*
  * For casting functions with int/void results
  */
 typedef void (* voidfunc) ();
+typedef void (* signalfunc) ();	/* (int) */
 typedef unsigned long (* charfunc) ();
 typedef int (* intfunc) ();
-
-/*
- * The regex structure. Status can be any of 0, BEGIN_LINE or REG_ERROR. In
- * the last case, the result.err_mess field is assigned. Start_ptr and end_ptr
- * point to the match found. For more details see the documentation file.
- */
-struct regex {
-  union {
-  	char * err_mess;
-  	int * expression;
-  } result;
-  char status;
-  LINE * start_line;
-  char * start_ptr;
-  LINE * end_line;
-  char * end_ptr;
-};
-
-typedef struct regex REGEX;
+typedef int (* flagfunc) ();	/* (menuitemtype *, int) */
 
 /* NULL definitions */
 #define NIL_PTR		((char *) 0)
 #define NIL_LINE	((LINE *) 0)
-#define NIL_REG		((REGEX *) 0)
-#define NIL_INT		((int *) 0)
 
 
 /***************************************************************************\
@@ -445,7 +417,6 @@ extern char yankie_file [];	/* Temp. file for inter-window buffer */
 extern char html_file [];	/* Temp. file for HTML embedding buffer */
 extern char panic_file [];	/* file for panic-write-back */
 extern FLAG yank_status;	/* Status of yank_file */
-extern FLAG html_status;	/* Status of html_file */
 extern FLAG redraw_pending;	/* was a redraw suppressed in find_y ? */
 extern long chars_saved;	/* # of chars saved in paste buffer */
 extern long bytes_saved;	/* # of bytes saved in paste buffer */
@@ -465,15 +436,30 @@ extern FLAG emacs_buffer;	/* enable emacs buffer fct for ^K/^T */
 extern FLAG paste_stay_left;	/* cursor stays before pasted region */
 extern character control_prefix; /* ^V/^P/^Q character to prefix control chars */
 extern character quit_char;	/* ^\/^G character to cancel command */
-extern FLAG Turkish;		/* enable Turkish case toggle specials ? */
+extern FLAG Turkish;		/* use Turkish case toggle specials ? */
+extern FLAG Lithuanian;		/* use Lithuanian case toggle specials ? */
 extern FLAG smart_quotes;	/* replace " with typographic quote ? */
+
 extern FLAG utf8_screen;	/* screen driven in UTF-8 mode ? */
 extern FLAG utf8_input;		/* keyboard input in UTF-8 mode ? */
 extern FLAG cjk_term;		/* terminal in CJK mode ? */
+extern FLAG cjk_uni_term;	/* terminal in CJK mode with Unicode widths ? */
 extern FLAG gb18030_term;	/* does CJK terminal support GB18030 ? */
 extern FLAG euc3_term;		/* does CJK terminal support EUC 3 byte ? */
 extern FLAG euc4_term;		/* does CJK terminal support EUC 4 byte ? */
 extern FLAG cjklow_term;	/* does CJK terminal support 81-9F range ? */
+extern int cjk_tab_width;	/* width of CJK TAB indicator */
+extern int cjk_lineend_width;	/* width of CJK line end indicator */
+extern FLAG combining_screen;	/* UTF-8 combining character terminal ? */
+extern FLAG bidi_screen;	/* UTF-8 bidi terminal ? */
+extern int width_data_version;
+extern int nonbmp_width_data;
+extern int combining_data_version;
+extern FLAG suppress_unknown_cjk;	/* on CJK terminal if no Unicode mapping */
+extern FLAG suppress_extended_cjk;	/* on CJK terminal if in extended code range */
+extern FLAG suppress_invalid_cjk;	/* on CJK terminal if invalid CJK code */
+extern FLAG utf_cjk_wide_padding;	/* always display CJK on UTF double-width ? */
+
 extern FLAG cjk_text;		/* text in CJK encoding ? */
 extern FLAG utf8_text;		/* text in UTF-8 representation ? */
 extern FLAG mapped_text;	/* text in 8 bit, non-Latin-1 representation ? */
@@ -483,10 +469,6 @@ extern FLAG disp_scrollbar;	/* shall scrollbar be displayed ? */
 extern FLAG fine_scrollbar;	/* fine-grained UTF-8 scrollbar ? */
 extern FLAG update_scrollbar_lazy;	/* partial scrollbar refresh as needed ? */
 extern FLAG combining_mode;	/* UTF-8 combining character support ? */
-extern int width_data_version;
-extern FLAG utf_cjk_wide_padding; /* always display CJK on UTF double-width ? */
-extern FLAG combining_screen;	/* UTF-8 combining character terminal ? */
-extern FLAG bidi_screen;	/* UTF-8 bidi terminal ? */
 extern FLAG display_block_graphics;	/* display control chars as graphics */
 extern FLAG no_window_title;	/* suppress filename display in window title? */
 
@@ -517,6 +499,7 @@ extern char TABchar;		/* Char to be shown in place of tab chars */
 extern char TABchar0;		/* Char to be shown at start of tab chars */
 extern char TABchar2;		/* Char to be shown at end of tab chars */
 extern char TABcharmid;		/* Char to be shown in middle of tab chars */
+extern unsigned long CJK_TAB_MARK;	/* to be shown in place of tab */
 extern char SHIFT_BEG;		/* Char indicating that line continues left */
 extern char RET_MARK;		/* Char indicating end of line */
 extern char DOSRET_MARK;	/* Char indicating DOS end of line */
@@ -542,15 +525,28 @@ extern char SHIFT_MARK;		/* Char indicating that line continues */
 extern character ring;		/* Ring accent character */
 #define UNI_ring (character) ''
 #define UNI_acute (character) ''
-#ifdef pc_charset
 #define DOS_ring (character) 0xF8
 #define DOS_acute (character) 0xEF
+#ifdef pc_charset
+#define pc_term True
+#else
+extern FLAG pc_term;		/* PC (CP437) character set ? */
 #endif
 
-extern FLAG rpipe;		/* Set if file should be read from stdin */
-extern FLAG stat_visible;	/* Set if status_line is visible */
-extern FLAG top_line_scrolled;	/* was menu line scrolled away? */
+extern FLAG stat_visible;	/* Set if status line is visible */
+extern FLAG top_line_scrolled;	/* Was menu line scrolled away? */
+extern FLAG always_disp_fstat;	/* Permanent file status display on status line? */
 extern FLAG always_disp_code;	/* Permanent char code display on status line? */
+extern FLAG always_disp_Han;	/* Permanent Han character description display on status line? */
+extern FLAG disp_Han_Mandarin;	/* display this Han pronunciation ? */
+extern FLAG disp_Han_Cantonese;	/* display this Han pronunciation ? */
+extern FLAG disp_Han_Japanese;	/* display this Han pronunciation ? */
+extern FLAG disp_Han_Sino_Japanese;	/* display this Han pronunciation ? */
+extern FLAG disp_Han_Korean;	/* display this Han pronunciation ? */
+extern FLAG disp_Han_Vietnamese;	/* display this Han pronunciation ? */
+extern FLAG disp_Han_description;	/* display Han description ? */
+extern FLAG disp_Han_full;	/* display full Han description ? */
+
 extern FLAG use_extracset;	/* Set if MINEDMARK enables graph. mark mode */
 extern FLAG use_curses;		/* True if compiled with CURSES */
 
@@ -559,7 +555,7 @@ extern char * pc_charmap;	/* PC character set mapping */
 extern voidfunc key_map [256];
 extern voidfunc ws_key_map [32];
 extern voidfunc emacs_key_map [32];
-extern voidfunc pc_key_map [256];
+extern voidfunc pc_fkey_map [256];
 
 extern FLAG page_scroll;	/* use scroll for page up/down */
 extern FLAG page_stay;		/* stay at edge of page after page up/down */
@@ -577,77 +573,97 @@ extern FLAG can_alt_cset;
 	declaration of mined functions
 \***************************************************************************/
 
-extern int find_y ();
-extern int find_y_w_o_RD ();
-extern int yankfile ();
-extern long get_number ();
-extern int get_hex_number ();
-extern void display ();
-extern void reset ();
-extern void move_y ();
-extern void move_to ();
-extern void set_cursor_xy ();
-extern void put_line ();
-extern void move_address ();
-extern void move_address_w_o_RD ();
+/* text buffer navigation */
+/* from textbuf.c */
+extern LINE * proceed _((LINE *, int));
+extern void reset _((LINE * scr_top, int y));
+extern void move_y _((int y));
+extern void move_to _((int x, int y));
+extern void move_address _((char *, int y));
+extern void move_address_w_o_RD _((char *, int y));
+extern int find_x _((LINE * line, char * address));
+extern int get_cur_col _((void));
+/* from mined2.c */
+extern int find_y _((LINE * line));
+extern int find_y_w_o_RD _((LINE * line));
 
-extern int length_of _((char *));
-extern int utf8_col_count _((char *));
-extern int char_count _((char *));
-extern int utf8_count _((char *));
-extern int col_count _((char *));
+/* file handling */
+/* from textbuf.c */
+extern int get_line _((int fd, char * buffer, int * len));
+extern int panicwrite _((void));
 
-extern int insert ();
-extern void copy_string ();
-extern int get_file ();
-extern int get_line ();
-extern void bad_write ();
-extern void file_status ();
-extern void calc_line_no ();
-extern int find_x ();
-extern void RDwin ();
-extern void RD_window_title ();
-extern void clear_window_title ();
-extern void RD_y ();
-extern void panicio ();
-extern int panicwrite ();
-extern void edit_nth_file ();
-extern FLAG save_text_load_file ();
+/* text buffer status handling */
+/* from mined1.c */
+extern void set_modified _((void));
+/* from textbuf.c */
+extern FLAG checkoverwrite _((char *));
+extern void file_status _((char * message, long bytecount, long charcount, 
+			char * filename, int lines, FLAG textstat, 
+			FLAG writefl, FLAG changed, FLAG viewing));
+extern void calc_line_no _((void));
 
-extern void checkout (), checkin ();
+/* from mined2.c */
+extern void delete_yank_files _((void));
+extern int scratchfile _((FLAG read_write, FLAG append, char * buf_name, char * which, FLAG * buf_status));
+extern int yankfile _((FLAG read_write, FLAG append));
 
+
+/* editing functions */
+/* from mined2.c */
+extern void yank_HTML _((FLAG remove));
+extern void paste_HTML _((void));
+/* from minedaux.c */
+extern int delete_text _((LINE * startl, char * start, LINE * endl, char * end));
+extern LINE * line_insert _((LINE *, char *, int, lineend_type));
+extern int insert _((LINE *, char * location, char * string));
+/* from textbuf.c */
+extern void edit_nth_file _((int));
+extern FLAG save_text_load_file _((char * filename));
 extern lineend_type extract_lineend_type _((char *, int));
 extern lineend_type default_lineend;
 
-extern unsigned long compose ();
-extern unsigned long compose_mnemo ();
-extern int idfchar ();
-extern unsigned long grave ();
-extern unsigned long circumflex ();
-extern unsigned long acute ();
-extern unsigned long diaeresis ();
-extern unsigned long tilde ();
-extern unsigned long angstrom ();
-extern void insert_grave ();
-extern void insert_circumflex ();
-extern void insert_acute ();
-extern void insert_diaeresis ();
-extern void insert_tilde ();
-extern void insert_angstrom ();
+/* character input handling */
+/* from mined2.c */
+extern void insert_grave _((void));
+extern void insert_circumflex _((void));
+extern void insert_acute _((void));
+extern void insert_diaeresis _((void));
+extern void insert_tilde _((void));
+extern void insert_angstrom _((void));
+extern void shift_0 _((void));
+extern void shift_1 _((void));
+extern void shift_2 _((void));
+extern void shift_3 _((void));
+extern void shift_4 _((void));
+extern void shift_5 _((void));
+extern void shift_6 _((void));
+extern void shift_7 _((void));
+extern void shift_8 _((void));
+extern void shift_9 _((void));
 
-/* aux: */
-extern void delete_yank_files ();
-extern void catch_interrupt ();
-extern char * scan_int _((char *, int *));
+/* from keydefs.c */
+extern unsigned long compose _((unsigned long, unsigned long));
+extern unsigned long compose_mnemonic _((char *));
+extern unsigned long grave _((unsigned long));
+extern unsigned long circumflex _((unsigned long));
+extern unsigned long acute _((unsigned long));
+extern unsigned long diaeresis _((unsigned long));
+extern unsigned long tilde _((unsigned long));
+extern unsigned long angstrom _((unsigned long));
 
 /* status line handling */
-extern void rd_bottom_line ();
-extern void redraw_prompt ();
-extern void clear_lastline ();
+/* from prompt.c */
+extern void rd_bottom_line _((void));
+extern void redraw_prompt _((void));
+extern void clear_lastline _((void));
+extern int get_filename _((char * message, char * filename));
 
 
 /* keyboard input */
 extern voidfunc keyproc;	/* function addressed by entered function key */
+extern int ansi_params;
+extern int ansi_param [];
+extern char ansi_fini;
 extern char keyshift;		/* shift state of entered function key */
 /* shift state indications of escape sequences */
 /*
@@ -688,124 +704,238 @@ extern int tabsize;
 #define set_cursor(nx, ny)	move_cursor(nx, ny)
 
 
-extern void reset_get_line ();
-extern void show_get_l_errors ();
-extern char * serror ();
-extern char * serrorof ();
-extern int geterrno ();
+/* from textbuf.c */
+extern void reset_get_line _((void));
+extern void show_get_l_errors _((void));
 
-extern int get_idf ();
+/* from mined2.c */
+extern int get_idf _((char * idf_buf, char * text, char * start_line));
 
-extern void set_modified ();
+/* from search.c */
+extern void search_for _((char *, FLAG));
+extern void search_expr _((char *, FLAG));
+extern void search_corresponding _((char *, FLAG, char *));
+/* from mined2.c */
+extern void viewonlyerr _((void));
+extern void restrictederr _((void));
 
-extern LINE * proceed ();
-extern LINE * match ();
-extern LINE * line_insert ();
-extern char syntax_state ();
-extern void update_text_state ();
-extern char * num_out ();
-extern character promptyn ();
-extern FLAG display_scrollbar ();
-extern void scrollbar_scroll_up _((int));
-extern void scrollbar_scroll_down _((int));
-extern FLAG checkoverwrite ();
-extern void delete_file ();
-extern char * unnull ();
-extern void search_for ();
-extern void search_expr ();
-extern void search_corresponding ();
-extern void viewonlyerr ();
-extern void restrictederr ();
 
-/* editing functions */
-extern void yank_HTML ();
-extern void paste_HTML ();
-extern int delete_text ();
+/* file output functions */
+/* from textbuf.c */
+extern int writechar _((int fd, char));
+extern int write_lineend _((int fd, lineend_type return_type));
+
+/* special stuff */
+/* from minedaux.c */
+extern char * envvar _((char * name));
+extern void catch_interrupt _((int));
+extern void panic _((char *, char *));
+extern void panicio _((char * message, char * err));
+extern char * scan_int _((char *, int *));
+extern char * dec_out _((long));
+extern char * serror _((void));
+extern char * serrorof _((int));
+extern int geterrno _((void));
+extern char syntax_state _((char, char *));
+extern void update_text_state _((LINE *));
+extern void delete_file _((char *));
+extern char * unnull _((char *));
+extern void copy_string _((char * to, char * from));
+
+/* memory allocation */
+/* from minedaux.c */
+extern void * alloc _((int bytes));
+extern void free_space _((void *));
+extern LINE * alloc_header _((void));
+extern void free_header _((LINE *));
+
+/* command dispatch mapper */
+/* from mined1.c */
+extern voidfunc command _((unsigned long));
+
 
 /* prompt line */
+/* from prompt.c */
+extern character promptyn _((void));
 extern FLAG char_on_status_line; /* is output active on status line ? */
 extern FLAG input_active;
 extern int lpos;
+extern long get_number _((char * message, char firstdigit, int * result));
+extern int prompt_num_char _((unsigned long * result, unsigned long maxvalue));
+
+
+/* character properties */
+/* from mined2.c */
+extern char * script _((unsigned long));
+/* from charmaps.c */
+extern int iscontrol _((character));
+extern character controlchar _((character));
 
 /* output functions */
-extern int iscontrol ();
-extern character controlchar ();
-extern void putcharacter ();
-extern void putmark ();
-extern void put_blanks ();
-extern int flush_buffer ();
-
-/* screen output functions */
+/* from mined1.c */
+extern void RDwin _((void));
+extern void RD_window_title _((void));
+extern void clear_window_title _((void));
+extern void RD_y _((int y_pos));
+/* from output.c */
+extern void display _((int y, LINE *, int count, int new_pos));
+extern void bad_write _((int fd));
+/* from output.c */
+extern FLAG display_scrollbar _((FLAG update));
+extern void scrollbar_scroll_up _((int from));
+extern void scrollbar_scroll_down _((int from));
+/* from output.c */
+extern void putmark _((char mark, char * utfmark));
+extern void put_blanks _((int endpos));
+extern int flush_buffer _((int fd));
+#define flushout()	flush_buffer (output_fd)
+extern void set_cursor_xy _((void));
+extern void put_line _((int scr_y, LINE *, int offset, FLAG clear, FLAG prop_pos));
 #define putchar(c)	__putchar ((character) c)
-extern void put_utfchar ();
-extern int writeout ();
-extern void putoutstring ();
-extern void _put_unichar ();
-extern int _put_cjkchar ();
+extern void put1utfchar _((char *));
+/* put character values */
+extern void putcharacter _((character));
+extern void _put_unichar _((unsigned long));
+extern int _put_cjkchar _((unsigned long));
 /* insert cast needed for Turbo-C: */
 #define put_unichar(c)	_put_unichar ((unsigned long) c)
 #define put_cjkchar(c)	_put_cjkchar ((unsigned long) c)
+/* low-level functions */
+extern int putoutchar _((character));
+extern void putoutstring _((char *));
 
 /* character handling */
-extern char * script ();
+/* from minedaux.c */
+extern int length_of _((char *));
+/* from mined1.c */
+extern int utf8_col_count _((char *));
+extern int char_count _((char *));
+extern int utf8_count _((char *));
+extern int col_count _((char *));
+extern char * charbegin _((char *, char *));
+extern void advance_utf8_scr _((char * *, int *, char *));
+extern void advance_utf8 _((char * *));
+extern void advance_char_scr _((char * *, int *, char *));
+extern void advance_char _((char * *));
+extern void precede_char _((char * *, char *));
+extern unsigned long charvalue _((character *));
+extern unsigned long unicodevalue _((character *));
+extern unsigned long unicode _((unsigned long));
+extern unsigned long precedingchar _((char *, char *));
+extern int uniscrwidth _((unsigned long, char *, char *));
+extern int cjkscrwidth _((unsigned long, char *, char *));
+extern int UTF_len _((char));
+extern int CJK_len _((character *));
+extern int UTFseq_len _((char *));
+extern void utf8_info _((char *, int *, unsigned long *));
+extern int isjoined _((unsigned long, char *, char *));
+extern int iscombined _((unsigned long, char *, char *));
+#define multichar(c)	((character) c >= 0x80 && (cjk_text == False || cjk_encoding != 'S' || (character) c < 0xA1 || (character) c > 0xDF))
 
-/* file output functions */
-extern int writechar ();
-extern int write_lineend ();
+/* from iscombin.c, iswide.c, isrtol.c */
+extern int iscombining _((unsigned long));
+extern int iswide _((unsigned long));
+extern int is_right_to_left _((unsigned long ucs));
+/* from locale.c, encoding.c */
+extern int is_locale_utf8 _((void));
+extern char * locale_terminal_encoding _((void));
+extern char * locale_text_encoding _((void));
 
-/* special stuff */
-extern void panic ();
-extern char * alloc ();
-extern void free_space ();
-extern LINE * alloc_header ();
-extern void free_header ();
-
-extern voidfunc command ();
+/* from charmaps.c */
+extern int uniUTF _((unsigned long, character *));
+extern int cjkencode _((unsigned long, character *));
+extern char * encode_char _((unsigned long));
+extern FLAG valid_cjk _((unsigned long cjkchar, /* opt */ character * cjkbytes));
+extern unsigned long isolated_alef _((unsigned long));
 
 /* character encoding handling */
-extern unsigned long cjk ();
-extern unsigned int lookup_cjk ();
-extern void set_mapped_text ();
-extern unsigned long max_cjk_value ();
-
-/* charmaps exports */
-extern void set_cjk_table ();
-extern unsigned int map_char ();
-extern unsigned long unmap_char ();
+#define CHAR_UNKNOWN -2	/* unknown mnemonic */
+#define CHAR_INVALID -1	/* not mappable to encoding */
+#define no_char(c)	(c == CHAR_UNKNOWN || c == CHAR_INVALID)
+/* from charmaps.c */
+extern unsigned long cjk _((unsigned long));
+extern unsigned long lookup_cjk _((unsigned long));
+extern unsigned long max_char_value _((void));
+extern FLAG encoding_has_combining _((void));
+extern char get_char_encoding _((void));
+extern FLAG set_char_encoding _((char));
 extern char cjk_encoding;
 extern char * cjk_encoding_flag;
 
+
+/* Han character descriptions table */
+struct hanentry {
+	unsigned long unicode;
+	char * Mandarin;
+	char * Cantonese;
+	char * Japanese;
+	char * Sino_Japanese;
+	char * Korean;
+	char * Vietnamese;
+	char * Definition;
+};
+
+extern struct hanentry hantable [];
+extern unsigned int hantable_len;
+
+
+/* keyboard mapping tables */
+#define keymapping_stringtables
+#ifdef keymapping_stringtables
+#define keymaptabletype char *
+#else
+#define keymaptabletype struct keymap *
+struct keymap {
+	char * fk;
+	char * fp;
+};
+#endif
+
+extern struct keymapping {
+	keymaptabletype table;
+	char * shortcut;
+} keymappingtable [];
+extern unsigned int keymappingtable_len;
+
+
 /* keyboard mapping menu handling */
-#define allow_keymap ((utf8_input == True && (utf8_text == True || cjk_text == True || visciimode ())) || cjk_term == True)
 extern char selection_space;
+extern FLAG enforce_keymap;	/* enable keyboard mapping even on non-suitable terminal */
+#define allow_keymap ((utf8_input && (utf8_text || cjk_text || mapped_text)) || cjk_term || enforce_keymap)
 #define SPACE_NEXT	'n'
 #define SPACE_NEXTROW	'r'
 #define SPACE_SELECT	's'
 
 /* menu handling */
-extern void QUICKMEN (), openmenu (), openmenuat ();
-extern int popup_menu ();
-extern mousebutton mouse_button, mouse_lastbutton, mouse_prevbutton;
-extern int mouse_shift;
-extern int mouse_xpos, mouse_ypos, mouse_lastxpos, mouse_lastypos;
-extern FLAG report_release;
-extern void clean_menus ();
-extern int is_menu_open ();
-extern void displaymenuline ();
-extern void redrawmenu ();
-extern void displayflags ();
-extern void toggleappend ();
-extern void set_scripttag ();
-extern FLAG flags_changed;	/* should flags menu area be redrawn? */
-extern FLAG in_menu_border;
-
-/* menu data structure */
 typedef struct {
 	char * itemname;
 	voidfunc itemfu;
 	char * hopitemname;
+	flagfunc itemon;
 	char tag;
 	} menuitemtype;
+extern void separator _((void));
+extern mousebutton mouse_button, mouse_lastbutton, mouse_prevbutton;
+extern int mouse_shift;
+extern int mouse_xpos, mouse_ypos, mouse_lastxpos, mouse_lastypos;
+extern FLAG report_release;
+extern FLAG flags_changed;	/* Should flag menu area be redrawn? */
+extern FLAG in_menu_border;
+/* from mousemen.c */
+extern void QUICKMEN _((int column, int line));
+extern void openmenu _((int meni)), openmenuat _((int col));
+extern int popup_menu _((menuitemtype *, int menulen, int column, int line, 
+			char * title, FLAG disp_only, char * select_keys));
+extern void clean_menus _((void));
+extern int is_menu_open _((void));
+extern void displaymenuline _((void));
+extern void redrawmenu _((void));
+extern void displayflags _((void));
+extern void toggleappend _((void));
+extern void set_scripttag _((void));
+extern void handleKeymapmenu _((void));
+extern void handleQuotemenu _((void));
+extern void handleEncodingmenu _((void));
 
 /* from mined2.c */
 extern int quote_type;
@@ -824,13 +954,16 @@ extern void invoke_key_function _((unsigned long));
 
 extern void QUICKMENU _((void)), FILEMENU _((void)), EDITMENU _((void));
 extern void SEARCHMENU _((void)), EXTRAMENU _((void)), PARAMENU _((void));
+extern void handleFlagmenus _((void));
 extern void MOUSEescape _((void));
+extern void ANSIseq _((void));
 extern void MUP _((void)), MDN _((void)), MLF _((void)), MRT _((void)), MNW _((void)), MPW _((void));
 extern void ctrl_MLF _((void)), ctrl_MRT _((void));
-extern void BSEN _((void)), ESEN _((void)), SCORR _((void));
+extern void BSEN _((void)), ESEN _((void));
 extern void SD _((void)), SU _((void)), PD _((void)), PU _((void)), BFILE _((void)), EFILE _((void));
 extern void BLINE _((void)), ELINE _((void)), MPPARA _((void)), MNPARA _((void)), HIGH _((void)), LOW _((void));
-extern void SNL _((void)), LIB _((void)), DPC _((void)), DCC _((void)), DLN _((void)), DNW _((void)), DPW _((void));
+extern void SNL _((void)), LIB _((void)), Underline _((void));
+extern void DPC _((void)), DCC _((void)), DLN _((void)), DNW _((void)), DPW _((void));
 extern void DCC0 _((void)), DPC0 _((void));
 extern void CTRLINS _((void)), DLINE _((void)), TOGINS _((void)), ctrlQ _((void)), ctrlK _((void)), ctrlO _((void));
 extern void search_wrong_enc _((void));
@@ -844,23 +977,25 @@ extern void WB _((void)), INSFILE _((void));
 extern void HOMEkey _((void)), ENDkey _((void)), DELkey _((void)), smallHOMEkey _((void)), smallENDkey _((void));
 extern void HOP _((void)), CANCEL _((void)), ESCAPE _((void)), FUNKEY _((void));
 extern void EMAX _((void)), META _((void)), UNDO _((void));
-extern void CHDI _((void)), PBUF _((void)), CMD _((void)), SH _((void)), SUSP _((void));
+extern void CHDI _((void)), PRINT _((void)), CMD _((void)), SH _((void)), SUSP _((void));
 extern void LNCI _((void)), LNSW _((void));
 extern void LOWCAP _((void)), LOWER _((void)), UPPER _((void)), CAPWORD _((void));
 extern void Stag _((void));
 extern void SIDFW _((void)), SIDRV _((void)), HTML _((void)), MARKER _((void)), GOMARKER _((void));
-extern void EDITmode _((void)), VIEWmode _((void));
+extern void EDITmode _((void)), VIEWmode _((void)), toggleVIEWmode _((void));
 extern void NXTFILE _((void)), PRVFILE _((void)), NTHFILE _((void)), EXFILE _((void)), EXMINED _((void));
 extern void ADJLM _((void)), ADJFLM _((void)), ADJNLM _((void)), ADJRM _((void)), ADJPAGELEN _((void));
 extern void FS _((void)), FSTAT _((void)), HELPF _((void)), HELPS _((void)), HELPC _((void)), HELPA _((void));
 extern void display_code _((void)), display_the_code _((void));
+extern void display_Han _((char * cpoi, FLAG force_utf8));
 extern void toggletab _((void));
 extern void changeuni _((void)), changehex _((void)), changeoct _((void)), changedec _((void));
 extern void screensmaller _((void)), screenbigger _((void)), screenlesslines _((void)), screenmorelines _((void));
 
 extern void toggleKEYMAP _((void)), setupKEYMAP _((void));
+extern void toggle_encoding _((void));
 
-extern void Sdoublequote _((void)), Ssinglequote _((void));
+extern void Sdoublequote _((void)), Ssinglequote _((void)), Sdash _((void));
 extern void S _((character)), S0 _((character)), Scharacter _((unsigned long));
 
 extern void goline _((int)), goproz _((int));
@@ -868,39 +1003,12 @@ extern void goline _((int)), goproz _((int));
 extern void SIDF _((FLAG));
 extern void Pushmark _((void)), Popmark _((void));
 extern void MARKn _((int)), GOMAn _((int)), Markn _((int));
-extern void UML _((char)), REPT _((char)), BAD _((character));
+extern void UML _((char)), REPT _((char));
+extern void BADch _((unsigned long cmd));
+extern void SCORR _((FLAG pref_direction));
 
-/*
-   Recognize multi-character prefix bytes (CJK)
- */
-#define multichar(c)	((character) c >= 0x80 && (cjk_text == False || cjk_encoding != 'S' || (character) c < 0xA1 || (character) c > 0xDF))
+extern void checkout _((void)), checkin _((void));
 
-extern int visciimode _((void));
-extern int isviscii _((character));
-
-extern char * charbegin _((char *, char *));
-extern void advance_utf8_scr _((char * *, int *, char *));
-extern void advance_utf8 _((char * *));
-extern void advance_char_scr _((char * *, int *, char *));
-extern void advance_char _((char * *));
-extern void precede_char _((char * *, char *));
-extern unsigned long charvalue _((character *));
-extern unsigned long precedingchar _((char *, char *));
-extern int UTF_len _((char));
-extern int CJK_len _((character *));
-extern int UTFseq_len _((char *));
-extern int uniUTF _((unsigned long, character *));
-extern int cjkencode _((unsigned long, character *));
-extern void utf8_info _((char *, int *, unsigned long *));
-extern int uniscrwidth _((unsigned long, char *, char *));
-extern int iswide _((unsigned long));
-extern int iscombining _((unsigned long));
-extern int iscombined _((unsigned long, char *, char *));
-extern int isjoined _((unsigned long, char *, char *));
-extern unsigned long isolated_alef _((unsigned long));
-extern int is_locale_utf8 _((void));
-extern char * locale_terminal_encoding _((void));
-extern char * locale_text_encoding _((void));
 
 /*
  * String functions
@@ -921,15 +1029,17 @@ extern char * locale_text_encoding _((void));
 /*
  * Functions handling status_line. ON means in reverse video.
  */
-extern int bottom_line ();
+extern int bottom_line _((FLAG revfl, char *, char *, char * inbuf, FLAG statfl, char * term_input));
 #define status_msg(str)		status_line (str, NIL_PTR)
-#define status_line(str1, str2)		\
+#define status_line(str1, str2)	\
 	(void) bottom_line (ON, (str1), (str2), NIL_PTR, False, "")
-#define status_beg(str)			\
+#define status_beg(str)		\
 	(void) bottom_line (ON, (str), NIL_PTR, NIL_PTR, True, "")
-#define error(str1, str2)		\
+#define error(str1)		\
+	(void) bottom_line (ON, (str1), NIL_PTR, NIL_PTR, False, "")
+#define error2(str1, str2)	\
 	(void) bottom_line (ON, (str1), (str2), NIL_PTR, False, "")
-#define clear_status()			\
+#define clear_status()		\
 	(void) bottom_line (OFF, NIL_PTR, NIL_PTR, NIL_PTR, False, "")
 #define get_string(str1, str2, fl, term_chars)	\
 	bottom_line (ON, (str1), NIL_PTR, (str2), fl, term_chars)
@@ -938,13 +1048,16 @@ extern int bottom_line ();
  * Print info about current file and buffer.
  */
 #define fstatus(mess, bytes, chars)	\
-	file_status ((mess), (bytes), file_name, total_lines, True, writable, modified, viewonly, (chars))
+	file_status ((mess), (bytes), (chars), file_name, total_lines, True, writable, modified, viewonly)
 
 /*
  * Build formatted string.
  * If this definition is left out, a routine will be defined in mined1.c.
  */
-#define build_string sprintf /**/
+#define build_string sprintf
+#ifndef build_string
+extern void build_string ();
+#endif
 
 /*
  * Get real shift value.
