@@ -4,6 +4,7 @@
 \*======================================================================*/
 
 #include "mined.h"
+#include "io.h"
 
 int panic_level = 0;		/* To adjust error handling to situation */
 
@@ -12,6 +13,22 @@ int panic_level = 0;		/* To adjust error handling to situation */
 |*			Auxiliary routines				*|
 \*======================================================================*/
 
+/**
+   Return the value of an environment variable.
+   Do not return NULL.
+ */
+char *
+envvar (name)
+  char * name;
+{
+  char * val = getenv (name);
+  if (val) {
+	return val;
+  } else {
+	return "";
+  }
+}
+
 /*
  * Delete file.
  */
@@ -41,7 +58,7 @@ void
 panic_msg (msg)
   char * msg;
 {
-  if (isscreenmode == True) {
+  if (isscreenmode) {
 	status_msg (msg);
 	sleep (2);
   } else {
@@ -61,7 +78,7 @@ panicking (message, err, signum)
   panic_level ++;
 
   if (panic_level < 2) {
-	if (loading == False && modified == True) {
+	if (loading == False && modified) {
 		panic_written = panicwrite ();
 		if (panic_written == ERRORS) {
 			build_string (text_buffer, "Error writing panic file %s", panic_file);
@@ -77,7 +94,7 @@ panicking (message, err, signum)
 	} else if (err == NIL_PTR) {
 		build_string (text_buffer, "%s", message);
 	} else {
-		build_string (text_buffer, "%s (Error: %s)", message, err);
+		build_string (text_buffer, "%s (%s)", message, err);
 	}
 	panic_msg (text_buffer);
 
@@ -93,7 +110,7 @@ panicking (message, err, signum)
   }
 
   if (panic_level < 3) {
-	if (isscreenmode == True) {
+	if (isscreenmode) {
 		set_cursor (0, YMAX);
 		putchar ('\n');
 #ifdef unix
@@ -128,7 +145,7 @@ catch_interrupt (signum)
   int signum;
 {
   panicking ("External signal %d caught - terminating", NIL_PTR, signum);
-  catch_signals (catch_interrupt);
+  catch_signals ((signalfunc) catch_interrupt);
 }
 
 /*-------------------------------------------------------------------------*/
@@ -159,7 +176,7 @@ extern void free ();
 static int allocount = 0;
 #endif
 
-char *
+void *
 alloc (bytes)
   int bytes;
 {
@@ -182,7 +199,7 @@ alloc (bytes)
 
 void
 free_space (p)
-  char * p;
+  void * p;
 {
   freemem (p);
 }
@@ -205,7 +222,7 @@ alloc_headerblock (n)
   LINE * new_list;
   int i = 0;
 
-  new_list = (LINE *) alloc (n * blocksizeof (LINE));
+  new_list = alloc (n * blocksizeof (LINE));
   if (new_list == NIL_LINE) {
 	free_header_list = NIL_LINE;
   } else {
@@ -221,7 +238,7 @@ alloc_headerblock (n)
 LINE *
 alloc_header ()
 {
-/*  return (LINE *) alloc (sizeof (LINE)); */
+/*  return alloc (sizeof (LINE)); */
   LINE * new_header;
 
   if (free_header_list == NIL_LINE) {
@@ -272,31 +289,55 @@ unnull (s)
  * Output an (unsigned) long in a 10 digit field without leading zeros.
  * It returns a pointer to the first digit in the buffer.
  */
+static
 char *
-num_out (number)
+num_out (number, radix)
   long number;
+  long radix;
 {
   static char num_buf [11];		/* Buffer to build number */
   register long digit;			/* Next digit of number */
-  register long pow = 1000000000L;	/* Highest ten power of long */
+  register long pow;			/* Highest radix power of long */
   FLAG digit_seen = False;
   int i;
+  int maxdigs;
+  char * bufp = num_buf;
 
-  for (i = 0; i < 10; i ++) {
+  if (radix == 16) {
+	pow = 268435456L;
+	maxdigs = 8;
+  } else {
+	pow = 1000000000L;
+	maxdigs = 10;
+	if (number < 0) {
+		* bufp ++ = '-';
+		number = - number;
+	}
+  }
+
+  for (i = 0; i < maxdigs; i ++) {
 	digit = number / pow;		/* Get next digit */
 	if (digit == 0L && digit_seen == False && i != 9) {
-		num_buf [i] = ' ';
 	} else {
-		num_buf [i] = '0' + (char) digit;
+		if (digit > 9) {
+			* bufp ++ = 'A' + (char) (digit - 10);
+		} else {
+			* bufp ++ = '0' + (char) digit;
+		}
 		number -= digit * pow;	/* Erase digit */
 		digit_seen = True;
 	}
-	pow /= 10L;			/* Get next digit */
+	pow /= radix;			/* Get next digit */
   }
-  num_buf [11] = '\0';
-  for (i = 0; num_buf [i] == ' '; i ++)	/* Skip leading spaces */
-	;
-  return & num_buf [i];
+  * bufp = '\0';
+  return num_buf;
+}
+
+char *
+dec_out (number)
+  long number;
+{
+  return num_out (number, 10);
 }
 
 /*
@@ -304,6 +345,15 @@ num_out (number)
  * %s indicates a string argument, %d indicates an integer argument.
  */
 #ifndef build_string /* otherwise build_string is sprintf */
+
+static
+char *
+hex_out (number)
+  long number;
+{
+  return num_out (number, 16);
+}
+
 /* VARARGS */
 void
 build_string (buf, fmt, args)
@@ -314,10 +364,12 @@ build_string (buf, fmt, args)
   int * argptr = & args;
   char * scanp;
   FLAG islong;
+  int length = 0;
 
   while (* fmt) {
      if (* fmt == '%') {
 	fmt ++;
+	fmt = scan_int (fmt, & length);
 	if (* fmt == 'l') {
 		islong = True;
 		fmt ++;
@@ -329,17 +381,24 @@ build_string (buf, fmt, args)
 		scanp = (char *) * argptr;
 		break;
 	case 'd' :
-		if (islong == True) {
-			scanp = num_out ((long) * ((long *) argptr));
+		if (islong) {
+			scanp = dec_out ((long) * ((long *) argptr));
 			if (sizeof (long) > sizeof (int)) {
 				argptr ++;
 			}
 		} else {
-			scanp = num_out ((long) * argptr);
+			scanp = dec_out ((long) * argptr);
 		}
 		break;
 	case 'D' :
-		scanp = num_out ((long) * ((long *) argptr));
+		scanp = dec_out ((long) * ((long *) argptr));
+		if (sizeof (long) > sizeof (int)) {
+			argptr ++;
+		}
+		break;
+	case 'x' :
+	case 'X' :
+		scanp = hex_out ((long) * ((long *) argptr));
 		if (sizeof (long) > sizeof (int)) {
 			argptr ++;
 		}
@@ -347,7 +406,7 @@ build_string (buf, fmt, args)
 	default :
 		scanp = "";
 	}
-	while (* buf ++ = * scanp ++)
+	while ((* buf ++ = * scanp ++))
 		{}
 	buf --;
 	argptr ++;
@@ -357,6 +416,7 @@ build_string (buf, fmt, args)
   }
   * buf = '\0';
 }
+
 #endif /* ndef build_string */
 
 /*
@@ -364,9 +424,9 @@ build_string (buf, fmt, args)
  * returns the character pointer behind the last digit
  */
 char *
-scan_int (str, num)
+scan_int (str, nump)
   char * str;
-  int * num;
+  int * nump;
 {
   register char * chpoi = str;
   int negative = False;
@@ -379,23 +439,23 @@ scan_int (str, num)
 	chpoi ++;
   }
   if (* chpoi >= '0' && * chpoi <= '9') {
-	* num = 0;
+	* nump = 0;
   } else {
 	return str;
   }
   while (* chpoi >= '0' && * chpoi <= '9' && quit == False) {
-	* num *= 10;
-	* num += * chpoi - '0';
+	* nump *= 10;
+	* nump += * chpoi - '0';
 	chpoi ++;
   }
   if (negative) {
-	* num = - * num;
+	* nump = - * nump;
   }
   return chpoi;
 }
 
 /*
- * length_of () returns the number of bytes in the string `string'
+ * length_of () returns the number of bytes in the string 'string'
  * excluding the '\0'.
  */
 int
@@ -413,7 +473,7 @@ length_of (string)
 }
 
 /*
- * text_bytes_of () returns the number of bytes in the string `string'
+ * text_bytes_of () returns the number of bytes in the string 'string'
  * up to and excluding the first '\n'.
  */
 static
@@ -433,8 +493,8 @@ text_bytes_of (string)
 }
 
 /*
- * Copy_string () copies the string `from' into the string `to'. `To' must be
- * long enough to hold `from'.
+ * copy_string () copies the string 'from' into the string 'to' which 
+   must be long enough
  */
 void
 copy_string (to, from)
@@ -476,7 +536,7 @@ copy_string (to, from)
 # ifdef __CYGWIN__
 
 #include <errno.h>
-/* What's the purpose of this rotten renaming junk game ? */
+/* What's the purpose of this silly API renaming game ? */
 #define sys_nerr _sys_nerr
 #define sys_errlist _sys_errlist
 /*const char *const *sys_errlist = _sys_errlist;*/
@@ -495,7 +555,7 @@ copy_string (to, from)
 #   ifndef __USE_GNU
 /* Why did those Linux kernel hackers have to change the interface here, 
    inserting a completely superfluous "const"?
-   All this voguish "assumedly-modern C" junk only increases 
+   All this voguish "assumedly-modern C" crap only increases 
    portability problems and wastes developers' time. 
    It's nothing but a nuisance. */
   extern char * sys_errlist [];
@@ -658,13 +718,13 @@ make_line (buffer, length, return_type)
 
   if (new_line == NIL_LINE) {
 	ring_bell ();
-	error ("Cannot allocate more memory for new line header", NIL_PTR);
+	error ("Cannot allocate more memory for new line header");
 	return NIL_LINE;
   } else {
 	new_line->text = alloc (length + 1);
 	if (new_line->text == NIL_PTR) {
 		ring_bell ();
-		error ("Cannot allocate more memory for new line", NIL_PTR);
+		error ("Cannot allocate more memory for new line");
 		return NIL_LINE;
 	} else {
 		new_line->shift_count = 0;
@@ -678,7 +738,7 @@ make_line (buffer, length, return_type)
 }
 
 /*
- * Line_insert () inserts a new line with text pointed to by `string'.
+ * Line_insert () inserts a new line with text pointed to by 'string'.
  * It returns the address of the new line.
  */
 LINE *
@@ -710,7 +770,7 @@ line_insert (line, string, len, return_type)
 }
 
 /*
- * Insert () insert the string `string' at the given line and location.
+ * Insert () insert the string 'string' at the given line and location.
  */
 int
 insert (line, location, string)
@@ -724,17 +784,17 @@ insert (line, location, string)
   lineend_type old_return_type;
   lineend_type new_return_type;
 
-  if (viewonly == True) {
+  if (viewonly) {
 	viewonlyerr ();
 	return ERRORS;
   }
 
   if (length_of (textp) + text_bytes_of (string) >= MAX_CHARS) {
-	error ("Line too long", NIL_PTR);
+	error ("Line too long");
 	return ERRORS;
   }
 
-/* Copy part of line until `location' has been reached */
+/* Copy part of line until 'location' has been reached */
   while (textp != location) {
 	* bufp ++ = * textp ++;
   }
@@ -750,7 +810,7 @@ insert (line, location, string)
   newtext = alloc (length_of (text_buffer) + length_of (location) + 1);
   if (newtext == NIL_PTR) {
 	ring_bell ();
-	error ("Cannot allocate memory for insertion", NIL_PTR);
+	error ("Cannot allocate memory for insertion");
 	return ERRORS;
   } else { /* Install the new text in this line */
 	if (* (string - 1) == '\n') {		/* Insert a new line */
@@ -761,7 +821,7 @@ insert (line, location, string)
 		/*	line->return_type = top_line->return_type;	*/
 			line->return_type = default_lineend;
 			new_return_type = old_return_type;
-		} else if (utf8_lineends == True
+		} else if (utf8_lineends
 			   && ((keyshift & ctrlshift_mask) || (hop_flag > 0))) {
 			if (keyshift & ctrl_mask) {
 				line->return_type = lineend_LS;
@@ -854,29 +914,28 @@ delete_text (start_line, start_textp, end_line, end_textp)
   int newpos_offset = start_textp - textp;
   FLAG isdeleting_lastcombining = False;
   int redraw_cols = 0;
-  int utfcount;
   unsigned long unichar;
   char * cp;
 
-  if (viewonly == True) {
+  if (viewonly) {
 	viewonlyerr ();
 	return ret;
   }
 
   set_modified ();			/* File has been modified */
 
-  if (utf8_text == True && combining_mode == True) {
-	utf8_info (start_textp, & utfcount, & unichar);
+  if (combining_mode && encoding_has_combining ()) {
+	unichar = unicodevalue (start_textp);
 	if (iscombined (unichar, start_textp, start_line->text)) {
 		cp = start_textp;
-		advance_utf8 (& cp);
-		utf8_info (cp, & utfcount, & unichar);
+		advance_char (& cp);
+		unichar = unicodevalue (cp);
 		if (! iscombining (unichar)) {
 			isdeleting_lastcombining = True;
 			cp = start_textp;
 			do {
 				precede_char (& cp, start_line->text);
-				utf8_info (cp, & utfcount, & unichar);
+				unichar = unicodevalue (cp);
 			} while (cp != start_line->text && iscombining (unichar));
 			if (iswide (unichar)) {
 				redraw_cols = 2;
@@ -895,7 +954,7 @@ delete_text (start_line, start_textp, end_line, end_textp)
 
 /* Check if line doesn't exceed MAX_CHARS */
   if (count + length_of (end_textp) >= MAX_CHARS) {
-	error ("Line too long", NIL_PTR);
+	error ("Line too long");
 	return ret;
   }
 
@@ -921,7 +980,7 @@ delete_text (start_line, start_textp, end_line, end_textp)
 	newtext = alloc (length_of (text_buffer) + 1);
 	if (newtext == NIL_PTR) {
 		ring_bell ();
-		error ("No more memory after deletion", NIL_PTR);
+		error ("No more memory after deletion");
 		ret = ERRORS;
 	} else {
 		free_space (start_line->text);
@@ -939,8 +998,8 @@ delete_text (start_line, start_textp, end_line, end_textp)
 		move_to (nx, y);	/* Reset cur_text */
 	} else {			/* Just display last part of line */
 		move_address (cur_line->text + newpos_offset, y);
-		if (isdeleting_lastcombining == True) {
-			if (redraw_cols == 0 || proportional == True) {
+		if (isdeleting_lastcombining) {
+			if (redraw_cols == 0 || proportional) {
 				set_cursor (0, y);
 				line_print (y, start_line);
 			} else {
@@ -957,7 +1016,7 @@ delete_text (start_line, start_textp, end_line, end_textp)
 
   shift = last_y;	   /* Save value */
   reset (top_line, y);
-  if ((line_cnt <= SCREENMAX - y) && can_delete_line == True) {
+  if ((line_cnt <= SCREENMAX - y) && can_delete_line) {
 	clear_status ();
 	display (y, start_line, 0, y);
 	line = proceed (start_line, SCREENMAX - y - line_cnt + 1);
