@@ -5,10 +5,13 @@
 #############################################################################
 # compilation source options
 
-# one of the following options may be included in the make target for a
-# specific source file depending on the level of code revision
-PROTOFLAGS=	# -Wmissing-prototypes # -Wstrict-prototypes
-DECLFLAGS=	# -Wmissing-declarations
+# code checking
+# gcc only: checking for function signatures
+PROTOFLAGS=
+
+# suppress prototype stuff for the sake of compilers and/or include files 
+# that cannot handle it; keep that as a gcc-only feature for source checking
+CCFLAGS	= -DNOPROTO
 
 
 #############################################################################
@@ -25,7 +28,7 @@ DEBUG	=
 # compilation setup
 
 # Collection of compilation parameters:
-CFLAGS	= $(SYSFLAGS) $(OPT) $(DEBUG)
+CFLAGS	= $(SYSFLAGS) $(CCFLAGS) $(OPT) $(DEBUG)
 
 
 #############################################################################
@@ -34,5 +37,19 @@ CFLAGS	= $(SYSFLAGS) $(OPT) $(DEBUG)
 LINKOPTS=$(LDOPTS)
 
 
+#############################################################################
+# dynamic make targets
+# this version for non-GNU make
+
+# used for linking; the dependency is resolved by a generated makefile
+CHARMAPS=charmaps/*.o
+
+# this dependency is ensured by a dynamically generated makefile
+KEYMAPS=
+
+# explicit make targets for dynamically .generating and invoking makefiles
+MAKEMAPS=mkcharmaps mkkeymaps
+
+
 #############################################################################
 # end
