@@ -4,6 +4,7 @@
 \*======================================================================*/
 
 #include "mined.h"
+#include "io.h"
 
 
 FLAG char_on_status_line = False; /* is output active on status line ? */
@@ -30,6 +31,28 @@ static FLAG pagewrapped = False;
 static int bottom_lines = 0;
 
 
+/**
+   Print line-end mark.
+ */
+static
+void
+print_mark ()
+{
+	reverse_off ();
+	dim_on ();
+	if (cjk_term) {
+		put_cjkchar (cjk (0x300A));	/* 《 */
+		if (cjk_lineend_width == 2) {
+			lpos ++;
+		}
+	} else {
+		put_unichar ((character) RET_MARK);
+	}
+	lpos ++;
+	dim_off ();
+	reverse_on ();
+}
+
 /**
    Print a byte to status line.
    Account for column position and possible line-wrap.
@@ -37,7 +60,7 @@ static int bottom_lines = 0;
 static
 void
 print_byte (c)
-  register character c;
+  character c;
 {
   char_on_status_line = True;
 
@@ -48,7 +71,9 @@ print_byte (c)
 	bottom_lines ++;
   }
 
-  if (iscontrol (c)) {
+  if (c == '\n') {
+	print_mark ();
+  } else if (iscontrol (c)) {
 	putchar ('^');
 	lpos ++;
 	put_unichar (controlchar (c));
@@ -61,8 +86,119 @@ print_byte (c)
   char_on_status_line = False;
 }
 
+/**
+   Print a mapped character to status line.
+   Account for column position and possible line-wrap.
+   Only called if (mapped_text && combining_mode == False && encoding_has_combining ())
+ */
+static
+void
+print_mapped (c, char_begin, line_begin)
+  character c;
+  char * char_begin;
+  char * line_begin;
+{
+  int width;
+  unsigned long unichar = lookup_cjk (c);
+
+  char_on_status_line = True;
+
+/*
+  This would be generic; for combining_mode == False it's just width = 1; ...
+
+  if (iscombined (unichar, char_begin, line_begin)) {
+	width = 0;
+  } else {
+	width = 1;
+  }
+*/
+  width = 1;
+
+  if (lpos > XBREAK_bottom - 2 + width) {
+	putstring ("\r\n");
+	lpos = 0;
+	pagewrapped = True;
+	bottom_lines ++;
+  }
+
+  if (c == '\n') {
+	print_mark ();
+  } else if (iscontrol (c)) {
+	putchar ('^');
+	lpos ++;
+	put_unichar (controlchar (c));
+	lpos ++;
+  } else {
+	if (combining_mode == False && combining_screen && iscombined (unichar, char_begin, line_begin)) {
+		/* enforce separated display */
+		putcharacter (' ');
+		putcharacter (c);
+	} else {
+		putcharacter (c);
+	}
+
+	lpos ++;
+  }
+
+  char_on_status_line = False;
+}
+
+/**
+   Print a CJK character to status line.
+   Line position and wrapping, as well as status line flag, are handled 
+   in input_cjk.
+ */
+static
+void
+print_cjk (c, width, char_begin, line_begin)
+  unsigned long c;
+  int width;
+  char * char_begin;
+  char * line_begin;
+{
+  unsigned long unichar = lookup_cjk (c);
+
+	if (encoding_has_combining () && iscombined (unichar, char_begin, line_begin)) {
+		if (combining_mode == False && combining_screen) {
+		/* enforce separated display */
+		/*	disp_separate_combining = True;	*/
+		/*	combiningdisp_on ();		*/
+
+			if (width == 2) {
+				put_cjkchar (cjk (0x3000));
+				put_cjkchar (c);
+			} else {
+				if (isjoined (unichar, char_begin, line_begin)) {
+					/* prevent terminal ligature by 
+					   substituting joining character 
+					   with its ISOLATED FORM */
+					unichar = isolated_alef (unichar);
+					c = cjk (unichar);
+					if (no_char (c)) {
+						put_cjkchar (cjk (0x2135));
+					} else {
+						put_cjkchar (c);
+					}
+				} else {
+					put_cjkchar (' ');
+					put_cjkchar (c);
+				}
+			}
+		} else if (combining_screen == False) {
+		/*	disp_separate_combining = True;	*/
+		/*	combiningdisp_on ();		*/
+			put_cjkchar (c);
+		} else {
+			put_cjkchar (c);
+		}
+	} else {
+		put_cjkchar (c);
+	}
+}
+
 /**
    Print a Unicode character to status line.
+   Handle line position and wrapping, as well as status line flag.
  */
 static
 void
@@ -74,27 +210,26 @@ print_unichar (unichar, width, char_begin, line_begin)
 {
 	char_on_status_line = True;
 
-	if (lpos > XBREAK_bottom - 1) {
+	if (width < 0) {
+		width = uniscrwidth (unichar, char_begin, line_begin);
+	}
+
+	if (lpos > XBREAK_bottom - 2 + width) {
 		putstring ("\r\n");
 		lpos = 0;
 		pagewrapped = True;
 		bottom_lines ++;
 	}
 
-	if (width < 0) {
-		width = uniscrwidth (unichar, char_begin, line_begin);
-	}
 	lpos += width;
 
 	if (iscombined (unichar, char_begin, line_begin)) {
-		if (combining_mode == False 
-			&& combining_screen == True)
-		{	/* enforce separated display */
+		if (combining_mode == False && combining_screen) {
+		/* enforce separated display */
 		/*	disp_separate_combining = True;	*/
 		/*	combiningdisp_on ();		*/
 
-			if (iswide (unichar))
-			{
+			if (iswide (unichar)) {
 				put_unichar (0x3000);
 				put_unichar (unichar);
 			} else {
@@ -133,6 +268,7 @@ printlim_string (text, limit)
 {
   int utfcount;
   unsigned long unichar;
+  unsigned long cjkchar;
   int charwidth;
   char * start = text;
 
@@ -142,13 +278,14 @@ printlim_string (text, limit)
 		diagdisp_off ();
 	}
 
-	if (utf8_text == True) {
+	if (utf8_text) {
 		utf8_info (text, & utfcount, & unichar);
 		charwidth = uniscrwidth (unichar, text, start);
 	} else if (iscontrol (* text)) {
 		charwidth = 2;
-	} else if (cjk_text == True && multichar (* text)) {
-		charwidth = 2;
+	} else if (cjk_text /* && multichar (* text) */) {
+		cjkchar = charvalue (text);
+		charwidth = cjkscrwidth (unichar, text, start);
 	} else {
 		charwidth = 1;
 	}
@@ -164,17 +301,21 @@ printlim_string (text, limit)
 		bottom_lines ++;
 	}
 
-	if (iscontrol (* text)) {
+	if (* text == '\n') {
+		print_mark ();
+		text ++;
+	} else if (iscontrol (* text)) {
 		lpos += 2;
 		putchar ('^');
 		put_unichar (controlchar (* text ++));
-	} else if (utf8_text == True) {
+	} else if (utf8_text) {
 		print_unichar (unichar, charwidth, text, start);
 		text += utfcount;
-	} else if (cjk_text == True && multichar (* text)) {
-		lpos += 2;
-		put_cjkchar (charvalue (text));
-		advance_char (& text);
+	} else if (cjk_text /* && multichar (* text) */) {
+		/*int prev_lpos = lpos;*/
+		advance_char_scr (& text, & lpos, start);
+		/*put_cjkchar (cjkchar, lpos - prev_lpos);*/
+		put_cjkchar (cjkchar);
 	} else {
 		lpos ++;
 		putcharacter (* text ++);
@@ -235,18 +376,20 @@ swallow_dummy_quit_char ()
 \*======================================================================*/
 
 /*
-   Read hex or octal character value within prompt line.
+   Read numeric character value within prompt line.
  */
 static
 unsigned long
-prompt_hex (maxvalue)
+get_num_char (result, maxvalue)
+  unsigned long * result;
   unsigned long maxvalue;
 {
-  register character c;
+  unsigned long c;
   unsigned long unichar = 0;
   int base = 16;
   int i = 0;
   unsigned long maxfill = 0;
+  FLAG uniinput = False;
 
   diagdisp_off ();
 
@@ -254,10 +397,13 @@ prompt_hex (maxvalue)
 	flush ();
 	c = readcharacter ();
 
-	if (c == '#' && i == 0) {
+	if (i == 0 && c == '#') {
 		base = 8;
-	} else if (c == '=' && i == 0) {
+	} else if (i == 0 && c == '=') {
 		base = 10;
+	} else if (i == 0 && (c == 'u' || c == 'U' || c == '+')) {
+		uniinput = True;
+		maxvalue = 0x10FFFF;
 	} else if ((c == '\b' || c == '\177') && i > 0) {
 		i --;
 		unichar /= base;
@@ -288,7 +434,7 @@ prompt_hex (maxvalue)
 		break;
 	}
   }
-  if (quit == True) {
+  if (quit) {
 	ring_bell ();
   }
   while (i > 0) {
@@ -297,7 +443,12 @@ prompt_hex (maxvalue)
   }
 
   diagdisp_on ();
-  return unichar;
+  if (uniinput && (cjk_text || mapped_text)) {
+	* result = cjk (unichar);
+  } else {
+	* result = unichar;
+  }
+  return c;
 }
 
 
@@ -312,13 +463,39 @@ input_byte (inbuf, ptr, c)
   character c;
 {
   if (c > '\0' && (ptr - inbuf) < maxLINE_LEN) {
+	if (mapped_text && combining_mode == False && encoding_has_combining ()) {
+		print_mapped (c, ptr, inbuf);
+	} else {
+		print_byte (c);
+	}
+	putstring (" \b");
 	* ptr ++ = c;
 	* ptr = '\0';
-	print_byte (c);
-	putstring (" \b");
   } else {
 	ring_bell ();
   }
+
+  if (mapped_text && combining_mode && encoding_has_combining ()) {
+	if (iscombining (lookup_cjk (c))) {
+		/* refresh display from base char, needed on xterm */
+		unsigned long uc;
+		char * refreshptr = ptr;
+		do {
+			precede_char (& refreshptr, inbuf);
+			c = charvalue (refreshptr);
+			uc = lookup_cjk (c);
+		} while (refreshptr != inbuf && iscombining (uc));
+		if (! iscombining (uc)) {
+			putstring ("\b");
+		}
+		/* Now refresh the whole combined character */
+		while (refreshptr < ptr) {
+			c = charvalue (refreshptr ++);
+			put_unichar (lookup_cjk (c));
+		}
+	}
+  }
+
   return ptr;
 }
 
@@ -334,48 +511,70 @@ input_cjk (inbuf, ptr, c)
   int scrlen;
   character * cp;
 
-	if (c < 0x80) {
-		return input_byte (inbuf, ptr, c);
-	}
-
 	len = cjkencode (c, cjkbytes);
 	if (len > 0 && (ptr - inbuf - 1 + len) < maxLINE_LEN) {
-		if (c >= 0x100) {
-			scrlen = 2;
-		} else {	/* Shift-JIS single byte */
-			scrlen = 1;
-		}
+		scrlen = cjkscrwidth (c, ptr, inbuf);
 
 		char_on_status_line = True;
+
 		if (lpos > XBREAK_bottom - scrlen) {
 			putstring ("\r\n");
 			lpos = 0;
 			pagewrapped = True;
 			bottom_lines ++;
 		}
+
+		print_cjk (c, scrlen, ptr, inbuf);
+
 		cp = cjkbytes;
 		while (* cp != '\0') {
 			* ptr ++ = * cp;
 			cp ++;
 		}
 		* ptr = '\0';
-		put_cjkchar (c);
+
 		lpos += scrlen;
+
 		char_on_status_line = False;
 		putstring (" \b");
 	} else {
 		ring_bell ();
 	}
+
+	if (combining_mode && encoding_has_combining ()) {
+		if (iscombining (lookup_cjk (c))) {
+			/* refresh display from base char, needed on xterm */
+			unsigned long uc;
+			char * refreshptr = ptr;
+			do {
+				precede_char (& refreshptr, inbuf);
+				c = charvalue (refreshptr);
+				uc = lookup_cjk (c);
+			} while (refreshptr != inbuf && iscombining (uc));
+			if (! iscombining (uc)) {
+				if (cjkscrwidth (c, refreshptr, inbuf) == 2) {
+					putstring ("\b\b");
+				} else {
+					putstring ("\b");
+				}
+			}
+			/* Now refresh the whole combined character */
+			while (refreshptr < ptr) {
+				c = charvalue (refreshptr);
+				put_cjkchar (c);
+				advance_char (& refreshptr);
+			}
+		}
+	}
+
 	return ptr;
 }
 
 static
 char *
-add_input_buf (ptr, c, char_begin, line_begin)
+add_input_buf (ptr, c)
   char * ptr;
   character c;
-  char * char_begin;
-  char * line_begin;
 {
   * ptr ++ = c;
   return ptr;
@@ -398,8 +597,8 @@ input_unicode (inbuf, ptr, unichar)
 		ring_bell ();
 	    } else {
 		print_unichar (unichar, -1, inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0xC0 | (unichar >> 6), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | (unichar & 0x3F), inputptr, inbuf);
+		ptr = add_input_buf (ptr, 0xC0 | (unichar >> 6));
+		ptr = add_input_buf (ptr, 0x80 | (unichar & 0x3F));
 		* ptr = '\0';
 	    }
 	} else if (unichar < 0x10000) {
@@ -407,9 +606,9 @@ input_unicode (inbuf, ptr, unichar)
 		ring_bell ();
 	    } else {
 		print_unichar (unichar, -1, inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0xE0 | (unichar >> 12), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 6) & 0x3F), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | (unichar & 0x3F), inputptr, inbuf);
+		ptr = add_input_buf (ptr, 0xE0 | (unichar >> 12));
+		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 6) & 0x3F));
+		ptr = add_input_buf (ptr, 0x80 | (unichar & 0x3F));
 		* ptr = '\0';
 	    }
 	} else if (unichar < 0x200000) {
@@ -417,10 +616,10 @@ input_unicode (inbuf, ptr, unichar)
 		ring_bell ();
 	    } else {
 		print_unichar (unichar, -1, inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0xF0 | (unichar >> 18), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 12) & 0x3F), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 6) & 0x3F), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | (unichar & 0x3F), inputptr, inbuf);
+		ptr = add_input_buf (ptr, 0xF0 | (unichar >> 18));
+		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 12) & 0x3F));
+		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 6) & 0x3F));
+		ptr = add_input_buf (ptr, 0x80 | (unichar & 0x3F));
 		* ptr = '\0';
 	    }
 	} else if (unichar < 0x4000000) {
@@ -428,11 +627,11 @@ input_unicode (inbuf, ptr, unichar)
 		ring_bell ();
 	    } else {
 		print_unichar (unichar, -1, inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0xF8 | (unichar >> 24), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 18) & 0x3F), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 12) & 0x3F), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 6) & 0x3F), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | (unichar & 0x3F), inputptr, inbuf);
+		ptr = add_input_buf (ptr, 0xF8 | (unichar >> 24));
+		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 18) & 0x3F));
+		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 12) & 0x3F));
+		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 6) & 0x3F));
+		ptr = add_input_buf (ptr, 0x80 | (unichar & 0x3F));
 		* ptr = '\0';
 	    }
 	} else if (unichar < 0x80000000) {
@@ -440,19 +639,19 @@ input_unicode (inbuf, ptr, unichar)
 		ring_bell ();
 	    } else {
 		print_unichar (unichar, -1, inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0xFC | (unichar >> 30), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 24) & 0x3F), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 18) & 0x3F), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 12) & 0x3F), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 6) & 0x3F), inputptr, inbuf);
-		ptr = add_input_buf (ptr, 0x80 | (unichar & 0x3F), inputptr, inbuf);
+		ptr = add_input_buf (ptr, 0xFC | (unichar >> 30));
+		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 24) & 0x3F));
+		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 18) & 0x3F));
+		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 12) & 0x3F));
+		ptr = add_input_buf (ptr, 0x80 | ((unichar >> 6) & 0x3F));
+		ptr = add_input_buf (ptr, 0x80 | (unichar & 0x3F));
 		* ptr = '\0';
 	    }
 	} else {
 		ring_bell ();
 	}
 
-	if (combining_mode == True) {
+	if (combining_mode) {
 		if (iscombining (unichar)) {
 			/* refresh display from base char, needed on xterm */
 			do {
@@ -466,6 +665,9 @@ input_unicode (inbuf, ptr, unichar)
 					putstring ("\b");
 				}
 			}
+			/* This also assumes utf8_screen == True 
+			   because otherwise combining_mode would be False
+			 */
 			putstring (inputptr);
 		}
 	}
@@ -479,11 +681,14 @@ char *
 input_character (inbuf, ptr, c)
   char * inbuf;
   char * ptr;
-  unsigned long c;
+  long c;
 {
-	if (utf8_text == True) {
+	if (no_char (c)) {
+		ring_bell ();
+		return ptr;
+	} else if (utf8_text) {
 		return input_unicode (inbuf, ptr, c);
-	} else if (cjk_text == True) {
+	} else if (cjk_text) {
 		return input_cjk (inbuf, ptr, c);
 	} else if (c < 0x100) {
 		return input_byte (inbuf, ptr, c);
@@ -502,7 +707,7 @@ redraw_prompt ()
 		bottom_lines --;
 	}
 	rd_bottom_line ();
-	if (can_clear_eol == True) {
+	if (can_clear_eol) {
 		clear_eol ();
 	}
 }
@@ -513,53 +718,47 @@ erase1 (inbuf, ptr)
   char * inbuf;
   char * ptr;
 {
-  unsigned long unichar;
-  int utfcount;
-
-	ptr --;
 	diagdisp_off ();
+	precede_char (& ptr, inbuf);
 	if (iscontrol (* ptr)) {
+	    if (* ptr == '\n' && (cjk_lineend_width == 1 || ! cjk_term)) {
+		/* single-width newline indication */
+		putstring (" \b\b \b");
+		lpos = lpos - 1;
+	    } else {
 		putstring (" \b\b\b  \b\b");
 		lpos = lpos - 2;
-	} else if (utf8_text == True && (* ptr & 0xC0) == 0x80) {
-		/* UTF-8 sequence byte found; look for start */
-		ptr ++;
-		precede_char (& ptr, inbuf);
-		utf8_info (ptr, & utfcount, & unichar);
-		while (combining_mode == True 
-			&& ptr != inbuf && iscombined (unichar, ptr, inbuf)) {
-			precede_char (& ptr, inbuf);
-			utf8_info (ptr, & utfcount, & unichar);
-		}
-
-		if (combining_mode == False 
-		    || ! iscombined (unichar, ptr, inbuf)) {
-			if (iswide (unichar)) {
-				putstring (" \b\b\b  \b\b");
-				lpos = lpos - 2;
-			} else {
-				putstring (" \b\b \b");
-				lpos = lpos - 1;
-			}
-		}
+	    }
 	} else {
-		ptr ++;
-		precede_char (& ptr, inbuf);
-		if (cjk_text == True && multichar (* ptr)) {
-			if (utf_cjk_wide_padding == False
-			    && cjk_term == False
-			    && ! iswide (lookup_cjk (charvalue (ptr))))
-			{
-				putstring (" \b\b \b");
-				lpos = lpos - 1;
-			} else {
-				putstring (" \b\b\b  \b\b");
-				lpos = lpos - 2;
-			}
-		} else {
-			putstring (" \b\b \b");
-			lpos = lpos - 1;
+	    unsigned long charval;
+	    int charwidth;
+
+	    if (combining_mode && encoding_has_combining ()) {
+		charval = charvalue (ptr);
+		while (ptr != inbuf && iscombined (unicode (charval), ptr, inbuf)) {
+			precede_char (& ptr, inbuf);
+			charval = charvalue (ptr);
 		}
+	    }
+
+	    if (utf8_text) {
+		charwidth = uniscrwidth (charval, ptr, inbuf);
+	    } else if (cjk_text) {
+		charwidth = cjkscrwidth (charval, ptr, inbuf);
+	    } else if (mapped_text && combining_mode && encoding_has_combining ()
+			&& iscombined (lookup_cjk (charval), ptr, inbuf)) {
+		charwidth = 0;
+	    } else {
+		charwidth = 1;
+	    }
+
+	    if (charwidth == 2) {
+		putstring (" \b\b\b  \b\b");
+		lpos = lpos - 2;
+	    } else if (charwidth == 1) {
+		putstring (" \b\b \b");
+		lpos = lpos - 1;
+	    }
 	}
 	diagdisp_on ();
 	putstring (" \b");
@@ -601,15 +800,23 @@ input (inbuf, clearfl, term_input)
 	redraw_prompt ();
      }
 
-     if (mnemo != NIL_PTR) {
-	diagdisp_off ();
-	putstring (" \b");
-	flush ();
-	c = readcharacter_mapping_valid (False, True);
-     } else {
-	flush ();
-	c = readcharacter_mapping_valid (True, True);
+     while (True) {
+	if (mnemo != NIL_PTR) {
+		diagdisp_off ();
+		putstring (" \b");
+		flush ();
+		c = readcharacter ();
+	} else {
+		flush ();
+		c = readcharacter_mapped ();
+	}
+	if (no_char (c)) {
+		ring_bell ();
+	} else {
+		break;
+	}
      }
+
      if (command (c) == SNL || command (c) == Popmark) {
 	c = '\r';
      }
@@ -622,14 +829,15 @@ input (inbuf, clearfl, term_input)
 
      if (mnemo != NIL_PTR &&
          (c == ' ' || c == '\n' || c == '\r' || c == quit_char || c == '\033')) {
-	mnemoc = compose_mnemo (mnemo);
+	mnemoc = compose_mnemonic (mnemo);
+
+	diagdisp_on ();
 
 	/* restore prompt display, remove mnemonic */
 	while ((char *) ptr > mnemo) {
 		ptr = erase1 (inbuf, ptr);
 	}
 
-	diagdisp_on ();
 	mnemo = NIL_PTR;
 
 	if (c != quit_char && c != '\033') {
@@ -664,8 +872,8 @@ input (inbuf, clearfl, term_input)
 
 	case '\n' :		/* End of input */
 	case '\r' :
-		/* If inbuf is empty clear status_line */
-		return (ptr == inbuf && clearfl == True) ?
+		/* If inbuf is empty clear status line */
+		return (ptr == inbuf && clearfl) ?
 			NO_INPUT : FINE;
 
 	case (character) FUNcmd :
@@ -718,20 +926,16 @@ input (inbuf, clearfl, term_input)
 		if (c == control_prefix /* ^V/^P/^Q */) {
 			c = readcharacter ();
 			if (
-#ifdef pc_charset
-			    ! (utf8_text == True && utf8_input == False) ?
-			    c == DOS_ring :
-#endif
-			    c == UNI_ring
+			    pc_term && ! (utf8_text && utf8_input == False)
+			    ? c == DOS_ring
+			    : c == UNI_ring
 			   )
 			{
 				ptr = input_character (inbuf, ptr, angstrom (readcharacter ()));
 			} else if (
-#ifdef pc_charset
-			    ! (utf8_text == True && utf8_input == False) ?
-			    c == DOS_acute :
-#endif
-			    c == UNI_acute
+			    pc_term && ! (utf8_text && utf8_input == False)
+			    ? c == DOS_acute
+			    : c == UNI_acute
 			   )
 			{
 				ptr = input_character (inbuf, ptr, acute (readcharacter ()));
@@ -758,35 +962,31 @@ input (inbuf, clearfl, term_input)
 				}
 			default:
 				if (c == '#') {
-					if (utf8_text == True) {
-						c = prompt_hex (0x7FFFFFFF);
-					} else if (cjk_text == True) {
-						c = prompt_hex (max_cjk_value ());
-					} else {
-						c = prompt_hex (0xFF);
-					}
-					if (quit == True) {
+				    unsigned long cx;
+				    do {
+					cx = get_num_char (& c, max_char_value ());
+					if (quit) {
 						quit = False;
-						break;
+					} else {
+						ptr = input_character (inbuf, ptr, c);
 					}
-				} else if (c == ' ' && utf8_text == True) {
+				    } while (cx == ' ');
+				} else if (c == ' ') {
 					/* mark mnemonic start position 
 					   in input buffer */
 					mnemo = (char *) ptr;
 					diagdisp_off ();
-					break;
 				} else if ( (c > ' ' && c <= 'Z')
 				            || (c > '^' && c != 0x7F) ) {
 					c = compose (c, readcharacter ());
-					if (c == quit_char) {
-						c = 0;
-					}
+					ptr = input_character (inbuf, ptr, c);
+				} else {
+					ptr = input_character (inbuf, ptr, c);
 				}
-
-				ptr = input_character (inbuf, ptr, c);
 			}
 
 		} else {
+			/* c != control_prefix */
 			ptr = input_character (inbuf, ptr, c);
 		}
      }
@@ -804,7 +1004,8 @@ input (inbuf, clearfl, term_input)
 status_line (str1, str2) is (void) bottom_line (ON, (str1), (str2), NIL_PTR, False, "")
 status_msg (str)	 is status_line (str, NIL_PTR)
 status_beg (str)	 is (void) bottom_line (ON, (str), NIL_PTR, NIL_PTR, True, "")
-error (str1, str2)	 is (void) bottom_line (ON, (str1), (str2), NIL_PTR, False, "")
+error (str1)		 is (void) bottom_line (ON, (str1), NIL_PTR, NIL_PTR, False, "")
+error2 (str1, str2)	 is (void) bottom_line (ON, (str1), (str2), NIL_PTR, False, "")
 clear_status ()		 is (void) bottom_line (OFF, NIL_PTR, NIL_PTR, NIL_PTR, False, "")
 get_string (str1, str2, fl, term_chars)
 			 is bottom_line (ON, (str1), NIL_PTR, (str2), fl, term_chars)
@@ -845,7 +1046,7 @@ bottom_line (revfl, s1, s2, inbuf, statfl, term_input)
   }
   lastinbuf = inbuf;
 
-  if (pagewrapped == True) {
+  if (pagewrapped) {
 	status_buf [0] = '\0';
 	RD ();
 	pagewrapped = False;
@@ -854,7 +1055,7 @@ bottom_line (revfl, s1, s2, inbuf, statfl, term_input)
   build_string (status_buf, " %s%s ", unnull (s1), unnull (s2));
 		/* (s1 == NIL_PTR) ? "" : s1, (s2 == NIL_PTR) ? "" : s2); */
 
-  if (revfl == ON && stat_visible == True) {
+  if (revfl == ON && stat_visible) {
 	set_cursor (0, YMAX);
 	clear_lastline ();
   }
@@ -881,11 +1082,11 @@ bottom_line (revfl, s1, s2, inbuf, statfl, term_input)
   /* Print normal video */
   diagdisp_off ();
   /* set_cursor (lpos, YMAX); not needed; wouldn't work proportional */
-  if (can_clear_eol == True) {
+  if (can_clear_eol) {
 	clear_eol ();
   } else {
 	put_blanks (XMAX - 1 - lpos);
-	if (proportional == True) {
+	if (proportional) {
 		set_cursor (0, YMAX);
 		diagdisp_on ();
 		if (lastinbuf == NIL_PTR) {
@@ -904,7 +1105,7 @@ bottom_line (revfl, s1, s2, inbuf, statfl, term_input)
 
   if (inbuf != NIL_PTR) {
 	set_cursor (0, YMAX);
-  } else if (statfl == True) {
+  } else if (statfl) {
 	diagdisp_on ();
   } else {
 	set_cursor_xy ();	/* Set cursor back to old position */
@@ -917,9 +1118,36 @@ bottom_line (revfl, s1, s2, inbuf, statfl, term_input)
   return ret;
 }
 
+/*
+ * Get_file () reads a filename from the terminal.
+ */
+int
+get_filename (message, file)
+  char * message;
+  char * file;
+{
+  int ret = get_string (message, file, True, "");
+#ifndef msdos
+  char * filei;
+  char file1 [maxLINE_LEN];
+
+  if (file [0] == '~' && file [1] == '/') {
+	filei = file; filei ++;
+	build_string (file1, "%s%s", envvar ("HOME"), filei);
+	build_string (file, file1);
+  }
+#endif
+  return ret;
+}
+
+
+/*======================================================================*\
+|*				Number input				*|
+\*======================================================================*/
+
 /*
  * get_number () reads a number from the terminal.
- * get_hex_number () reads a hexadecimal number from the terminal.
+ * prompt_num_char () prompts for a numeric character from the terminal.
  * If the parameter firstdigit is a valid digit, empty input (return) 
  * will yield a valid number; if it is '\0', it won't.
  * The last character typed in is returned.
@@ -932,34 +1160,34 @@ get_number (message, firstdigit, result)
   char firstdigit;
   int * result;
 {
-  register long index;
-  register int count;
-  register int i = 0;
+  long ch;
+  int value;
+  int i = 0;
 
   status_beg (message);
 
   if (firstdigit > '\0') {
-	index = firstdigit;
+	ch = firstdigit;
   } else {
-	index = readcharacter_mapping (True);
+	ch = readcharacter_mapped ();
   }
 
-  if (index == quit_char || index == '\033') {
+  if (ch == quit_char || ch == '\033') {
 	quit = True;
   }
-  if (quit == False && (index < '0' || index > '9')) {
-	error ("Bad number", NIL_PTR);
+  if (quit == False && (ch < '0' || ch > '9')) {
+	error ("Bad number");
 	return ERRORS;
   }
 
 /* Convert input to a decimal number */
-  count = 0;
-  while (quit == False && ((index >= '0' && index <= '9')
-			   || index == '\b' || index == '\177')) {
-	if (index == '\b' || index == '\177') {
+  value = 0;
+  while (quit == False && ((ch >= '0' && ch <= '9')
+			   || ch == '\b' || ch == '\177')) {
+	if (ch == '\b' || ch == '\177') {
 		if (i > 0) {
 			i --;
-			count /= 10;
+			value /= 10;
 			putstring ("\b \b"); flush ();
 			if (lpos >= XBREAK_bottom) {
 				pagewrapped = True;
@@ -968,40 +1196,41 @@ get_number (message, firstdigit, result)
 		}
 	} else {
 		i ++;
-		print_byte (index); flush ();
+		print_byte (ch); flush ();
 		if (lpos >= XBREAK_bottom) {
 			pagewrapped = True;
 		}
-		count *= 10;
-		count += index - '0';
+		value *= 10;
+		value += ch - '0';
 	}
-	index = readcharacter_mapping (True);
-	if (index == quit_char || index == '\033') {
+	ch = readcharacter_mapped ();
+	if (ch == quit_char || ch == '\033') {
 		quit = True;
 	}
   }
 
   clear_status ();
-  if (quit == True) {
+  if (quit) {
 	clear_status ();
 	return ERRORS;
   }
-  * result = count;
-  return index;
+  * result = value;
+  return ch;
 }
 
 /*
    Prompt for hex or octal character value.
  */
 int
-get_hex_number (result, maxvalue)
+prompt_num_char (result, maxvalue)
   unsigned long * result;
   unsigned long maxvalue;
 {
-  register int index;
-  register int count = 0;
+  unsigned long ch;
+  int value = 0;
   int i;
   FLAG ok = False;
+  FLAG uniinput = False;
   int base = 16;
   unsigned long maxfill = 0;
 
@@ -1009,40 +1238,47 @@ get_hex_number (result, maxvalue)
 
   i = 0;
   while (maxfill < maxvalue && ok == False) {
-	index = readcharacter ();
-	if (i == 0 && index == '#') {
-		base = 8;
-		print_string ("(oct) "); flush ();
-		index = readcharacter ();
-	} else if (i == 0 && index == '=') {
-		base = 10;
-		print_string ("(dec) "); flush ();
-		index = readcharacter ();
+	ch = readcharacter ();
+	if (i == 0) {
+		if (ch == '#') {
+			base = 8;
+			print_string ("(oct) "); flush ();
+			continue;
+		} else if (ch == '=') {
+			base = 10;
+			print_string ("(dec) "); flush ();
+			continue;
+		} else if (ch == 'u' || ch == 'U' || ch == '+') {
+			uniinput = True;
+			maxvalue = 0x10FFFF;
+			print_string ("(uni) "); flush ();
+			continue;
+		}
 	}
 
-	if (index >= '0' && index <= '9' && (index <= '7' || base > 8)) {
+	if (ch >= '0' && ch <= '9' && (ch <= '7' || base > 8)) {
 		i ++;
-		count *= base;
-		count += index - '0';
+		value *= base;
+		value += ch - '0';
 		maxfill = maxfill * base + base - 1;
-		print_byte (index); flush ();
-	} else if (base == 16 && index >= 'A' && index <= 'F') {
+		print_byte (ch); flush ();
+	} else if (base == 16 && ch >= 'A' && ch <= 'F') {
 		i ++;
-		count *= base;
-		count += index - 'A' + 10;
+		value *= base;
+		value += ch - 'A' + 10;
 		maxfill = maxfill * base + base - 1;
-		print_byte (index); flush ();
-	} else if (base == 16 && index >= 'a' && index <= 'f') {
+		print_byte (ch); flush ();
+	} else if (base == 16 && ch >= 'a' && ch <= 'f') {
 		i ++;
-		count *= base;
-		count += index - 'a' + 10;
+		value *= base;
+		value += ch - 'a' + 10;
 		maxfill = maxfill * base + base - 1;
-		index = index - 'a' + 'A';
-		print_byte (index); flush ();
-	} else if (index == '\b' || index == '\177') {
+		ch = ch - 'a' + 'A';
+		print_byte (ch); flush ();
+	} else if (ch == '\b' || ch == '\177') {
 		if (i > 0) {
 			i --;
-			count /= base;
+			value /= base;
 			maxfill /= base;
 			putstring ("\b \b"); flush ();
 			if (lpos >= XBREAK_bottom) {
@@ -1050,7 +1286,14 @@ get_hex_number (result, maxvalue)
 			}
 			lpos --;
 		}
-	} else if (index == ' ' || index == '\n' || index == '\r') {
+	} else if (ch == '\033' || ch == quit_char) {
+		clear_status ();
+		return ERRORS;
+	} else if (ch == ' ' || ch == '\n' || ch == '\r') {
+		if (i == 0) {
+			clear_status ();
+			return ERRORS;
+		}
 		ok = True;
 	} else {
 		quit = True;
@@ -1063,41 +1306,24 @@ get_hex_number (result, maxvalue)
   }
 
   clear_status ();
-  if (quit == True) {
+  if (quit) {
 	if (base == 16) {
-		error ("Bad hex value", NIL_PTR);
+		error ("Bad hex value");
 	} else if (base == 8) {
-		error ("Bad octal value", NIL_PTR);
+		error ("Bad octal value");
 	} else {
-		error ("Bad decimal value", NIL_PTR);
+		error ("Bad decimal value");
 	}
 	quit = False;
 	return ERRORS;
   }
-  * result = count;
-  return index;
-}
 
-/*
- * Get_file () reads a filename from the terminal.
- */
-int
-get_file (message, file)
-  char * message;
-  char * file;
-{
-  int ret = get_string (message, file, True, "");
-#ifndef msdos
-  char * filei;
-  char file1 [maxLINE_LEN];
-
-  if (file [0] == '~' && file [1] == '/') {
-	filei = file; filei ++;
-	build_string (file1, "%s%s", unnull (getenv ("HOME")), filei);
-	build_string (file, file1);
+  if (uniinput && (cjk_text || mapped_text)) {
+	* result = cjk (value);
+  } else {
+	* result = value;
   }
-#endif
-  return ret;
+  return ch;
 }
 
 
