@@ -4,23 +4,57 @@
 \*======================================================================*/
 
 #include "mined.h"
+#include "io.h"
+
+
+/* Expression definitions */
+#define NO_MATCH	0
+#define MATCH		1
+#define REG_ERROR	2
+
+#define BEGIN_LINE	(2 * REG_ERROR)
+#define END_LINE	(2 * BEGIN_LINE)
+
+/**
+   The regex structure. Status can be any of 0, BEGIN_LINE or REG_ERROR.
+   In the last case, the result.err_mess field is assigned.
+   Start_ptr and end_ptr point to the match found.
+   For more details see the documentation file.
+ */
+struct regex {
+  union {
+  	char * err_mess;
+  	int * expression;
+  } result;
+  char status;
+  LINE * start_line;
+  char * start_ptr;
+  LINE * end_line;
+  char * end_ptr;
+};
+
+typedef struct regex REGEX;
+
+#define NIL_REG		((REGEX *) 0)
+#define NIL_INT		((int *) 0)
 
-/* external */
-extern FLAG redraw_pending;	/* was a redraw suppressed in find_y ? */
 
 /* forward */
-static char * substitute ();
-static int compile ();
-static void search ();
-static void re_search ();
-static void prev_search ();
-static void do_search ();
-static int line_check ();
-static int check_string ();
-static int in_exprlist ();
-static int star ();
-static LINE * search_it ();
-static REGEX * make_expression ();
+static char * substitute _((LINE * line, LINE * * line_poi, REGEX * program, 
+	int * old_lines_deleted, char * replacement, int * new_lines_inserted));
+static int compile _((character * pat_start, REGEX * program));
+static void search _((char * message, FLAG method));
+static void re_search _((void));
+static void prev_search _((void));
+static void do_search _((REGEX * program, FLAG method));
+static LINE * match _((REGEX * program, character * string, FLAG method));
+static int line_check _((REGEX * program, char * str, LINE * from, FLAG method));
+static int check_string _((REGEX * program, character * str, LINE * from, int * expression));
+static int in_exprlist _((int * exprlist, character * s, int exprlist_length, int opcode));
+static int star _((REGEX * program, char * end_position, char * str, LINE * from, int * expression));
+static LINE * search_it _((REGEX * program, FLAG method));
+static REGEX * make_expression _((char * expr));
+
 
 /* local */
 static FLAG wrapped_search;		/* temporary flag of wrapped search */
@@ -54,7 +88,7 @@ static FLAG prevmethod = NOT_VALID;	/* FORWARD, REVERSE */
  *	   that expression.
  */
 
-char typed_expression [maxLINE_LEN];	/* Holds previous search expression */
+static char typed_expression [maxLINE_LEN];	/* Holds previous search expression */
 
 /*
  * SIDFW searches forward for the current identifier.
@@ -81,6 +115,8 @@ search_cur_char ()
   char cbuf [7];
   char * cp = cur_text;
 
+  Pushmark ();
+
   advance_char (& cp);
   cbuf [0] = '\0';
   if (cp - cur_text > 6) {
@@ -143,17 +179,47 @@ SRV ()
   }
 }
 
+/**
+   reverse_method maps search directions FORWARD / REVERSE to each other,
+   leaving NOT_VALID as is
+ */
+static
+FLAG
+reverse_method (dir)
+  FLAG dir;
+{
+  if (dir == REVERSE) {
+	return FORWARD;
+  } else if (dir == FORWARD) {
+	return REVERSE;
+  } else {
+	return dir;
+  }
+}
+
 /*
  * RS searches using the last search direction and expression.
  */
 void
 RS ()
 {
-  if (keyshift & shift_mask) {
+  if ((keyshift & altshift_mask) == altshift_mask) {
+	keyshift = '0';
+	SIDF (REVERSE);
+  } else if (keyshift & shift_mask) {
 	keyshift = '0';
 	SIDFW ();
+  } else if (keyshift & alt_mask) {
+	keyshift = '0';
+	if (hop_flag > 0) {
+		hop_flag = 0;
+		search_for (prevexpr, reverse_method (prevmethod));
+	} else {
+		do_search (lastprogram, reverse_method (lastmethod));
+	}
   } else {
 	if (hop_flag > 0) {
+		hop_flag = 0;
 		prev_search ();
 	} else {
 		re_search ();
@@ -181,7 +247,7 @@ get_expression (message)
   }
 
   if (exp_buf [0] == '\0' && typed_expression [0] == '\0') {
-	error ("No previous search expression", NIL_PTR);
+	error ("No previous search expression");
 	return NIL_REG;
   }
 
@@ -194,7 +260,7 @@ get_expression (message)
   }
 
   if (program.status == REG_ERROR) {	/* Error during compiling */
-	error (program.result.err_mess, NIL_PTR);
+	error (program.result.err_mess);
 	return NIL_REG;
   }
 
@@ -218,7 +284,7 @@ make_expression (expr)
   }
 
   if (program.status == REG_ERROR) {	/* Error during compiling */
-	error (program.result.err_mess, NIL_PTR);
+	error (program.result.err_mess);
 	return NIL_REG;
   }
   return & program;
@@ -228,8 +294,8 @@ make_expression (expr)
  * Change () prompts for an expression and a substitution pattern and changes
  * all matches of the expression into the substitution.
  * change () starts looking for expressions at the current line and
- * continues until the end of the file if the FLAG `global' is VALID.
- * It prompts for each change if the FLAG `confirm' is True.
+ * continues until the end of the file if the FLAG 'global' is VALID.
+ * It prompts for each change if the FLAG 'confirm' is True.
  * For a call graph of search procedures see search ().
  */
 static
@@ -256,7 +322,7 @@ change (message, global, confirm)
   int new_lines_inserted;
   int lines_changed = 0;
 
-  if (viewonly == True) {
+  if (viewonly) {
 	viewonlyerr ();
 	return;
   }
@@ -288,6 +354,10 @@ change (message, global, confirm)
   }
 */
 
+  if (confirm) {
+	Pushmark ();
+  }
+
   set_cursor (0, YMAX);
   flush ();
 /* Substitute until end of file */
@@ -296,7 +366,7 @@ change (message, global, confirm)
 	    lines ++;
 	    /* Repeat sub. on this line as long as we find a match */
 	    do {
-		if (confirm == True) {
+		if (confirm) {
 			ly = find_y (line);
 			textp = program->start_ptr;
 			move_address (program->start_ptr, ly);
@@ -385,25 +455,25 @@ change (message, global, confirm)
   quit_change = quit;
 /* Fix the status line */
   if (subs == 0L && quit == False) {
-	error ("Pattern not found", NIL_PTR);
-  } else if (lines >= REPORT || quit == True) {
+	error ("Pattern not found");
+  } else if (lines >= REPORT_CHANGED_LINES || quit) {
 	build_string (mess_buf, "%s %ld substitutions on %ld lines",
-		(quit_change == True) ? "(Aborted) " : "", subs, lines);
+		quit_change ? "(Aborted) " : "", subs, lines);
 	status_msg (mess_buf);
-  } else if (global == NOT_VALID && subs >= REPORT) {
-	status_line (num_out (subs), " substitutions");
+  } else if (global == NOT_VALID && subs >= REPORT_CHANGED_LINES) {
+	status_line (dec_out (subs), " substitutions");
   } else {
 	clear_status ();
   }
 
-  if (confirm == True) {
+  if (confirm) {
 	move_to (x, y);
   } else {
 /*	move_to (LINE_START, previousy);	*/
 	move_address (cur_line->text + text_offset, previousy);
   }
 
-/*  if (quit == True) swallow_dummy_quit_char (); */
+/*  if (quit) swallow_dummy_quit_char (); */
   quit = False;
 }
 
@@ -488,9 +558,9 @@ substitute (line, line_poi,
 /* Check for line length not exceeding MAX_CHARS */
   if (length_of (text_buffer) + length_of (program->end_ptr) >= MAX_CHARS) {
 	if (program->end_line != program->start_line) {
-		error ("Substitution failed: replacement buffer exceeded", NIL_PTR);
+		error ("Substitution failed: replacement buffer exceeded");
 	} else {
-		error ("Substitution failed: resulted line too long", NIL_PTR);
+		error ("Substitution failed: resulted line too long");
 	}
 	return NIL_PTR;
   }
@@ -525,7 +595,7 @@ einfgen; Rckkehrparameter no_of_new_lines einfhren und setzen */
   newtext = alloc (repl_len + 1);
   if (newtext == NIL_PTR) {
 	ring_bell ();
-	error ("Substitution failed: cannot allocate more memory", NIL_PTR);
+	error ("Substitution failed: cannot allocate more memory");
 	return NIL_PTR;
   }
   else {
@@ -552,7 +622,7 @@ einfgen; Rckkehrparameter no_of_new_lines einfhren und setzen */
 	line = line_insert (line, repl_start, repl_len, line->return_type);
 	if (line == NIL_LINE) {
 		ring_bell ();
-		error ("Substitution failed: cannot allocate more lines", NIL_PTR);
+		error ("Substitution failed: cannot allocate more lines");
 		return NIL_PTR;
 	}
 	* line_poi = line;
@@ -695,7 +765,7 @@ search_expr (expr, method)
 /*
  * search_corresponding searches for a corresponding bracket of a 
  * bracket pair expression.
- * The parameter `method` must be valid.
+ * The parameter 'method' must be valid.
  */
 void
 search_corresponding (expr, method, corresponding)
@@ -718,7 +788,7 @@ search_corresponding (expr, method, corresponding)
   }
 
   redraw_pending = False;
-/* strcpy (pflags, redraw_pending == True ? "T" : "F"); */
+/* strcpy (pflags, redraw_pending ? "T" : "F"); */
   wrapped_search = False;
   do {
 	do {
@@ -726,7 +796,7 @@ search_corresponding (expr, method, corresponding)
 		if (match_line == NIL_LINE) {
 			return;
 		}
-		if (wrapped_search == True) {
+		if (wrapped_search) {
 			status_msg ("No corresponding bracket");
 			move_address (old_text, find_y (old_line));
 			return;
@@ -765,7 +835,7 @@ search_corresponding (expr, method, corresponding)
 	if (bracount >= 0) {
 	/* set position to match for next search: */
 		move_address_w_o_RD (program->start_ptr, find_y_w_o_RD (match_line));
-/* strcat (pflags, redraw_pending == True ? "T" : "F"); */
+/* strcat (pflags, redraw_pending ? "T" : "F"); */
 	}
 /*
 	build_string (text_buffer, "bracount = %d", bracount);
@@ -773,7 +843,7 @@ search_corresponding (expr, method, corresponding)
 	sleep (1);
 	break;
 */
-	if (quit == True) {
+	if (quit) {
 		status_msg ("Bracket search aborted");
 		quit = False;
 		move_address (old_text, find_y (old_line));
@@ -783,13 +853,13 @@ search_corresponding (expr, method, corresponding)
 
 /* set position to old to enable correct function of find_y: */
   move_address_w_o_RD (old_text, find_y_w_o_RD (old_line));
-/* strcat (pflags, redraw_pending == True ? "T" : "F"); */
+/* strcat (pflags, redraw_pending ? "T" : "F"); */
 
 /* position and re-display if necessary: */
   was_redraw_pending = redraw_pending;
   move_address (program->start_ptr, find_y (match_line));
-/* strcat (pflags, redraw_pending == True ? "T" : "F"); */
-  if (was_redraw_pending == True && redraw_pending == True) {
+/* strcat (pflags, redraw_pending ? "T" : "F"); */
+  if (was_redraw_pending && redraw_pending) {
 	RD ();
 /*	status_line ("pending redraw performed ", pflags);	*/
   }
@@ -812,7 +882,7 @@ search_it (program, method)
   flush ();
 /* Find the match */
   if ((match_line = match (program, cur_text, method)) == NIL_LINE) {
-	if (quit == True) {
+	if (quit) {
 		status_msg ("Aborted");
 /*		swallow_dummy_quit_char (); */
 		quit = False;
@@ -832,11 +902,11 @@ do_search (program, method)
   register LINE * match_line;
 
   if (method == NOT_VALID) {
-	error ("No previous search", NIL_PTR);
+	error ("No previous search");
 	return;
   }
   if (program == NIL_REG) {
-	error ("No previous search expression", NIL_PTR);
+	error ("No previous search expression");
 	return;
   }
 
@@ -848,8 +918,10 @@ do_search (program, method)
 }
 
 
-/* Now the search and replace utilities */
-/* ------------------------------------ */
+/*======================================================================*\
+|*			Search/replace utility functions		*|
+\*======================================================================*/
+
 /* Opcodes for characters */
 #define NORMAL		0x0200
 #define DOT		0x0400
@@ -864,24 +936,25 @@ do_search (program, method)
 #define OPCODE_MASK	0xFE00
 #define LENGTH_MASK	0x01FF
 
-/* Previous is the contents of the previous address (ptr) points to */
+/* Previous is the content of the previous address (ptr) points to */
 #define previous(ptr)		(* ((ptr) - 1))
 
 /* Buffer to store outcome of compilation */
-int exp_buffer [BLOCK_SIZE];
+#define MAX_EXP_SIZE	LENGTH_MASK	/* 511 */
+static int exp_buffer [MAX_EXP_SIZE];
 
 /* Errors often used */
-char * too_long = "Regular expression too long";
+static char * too_long = "Regular expression too long";
 
 /*
- * Reg_error () is called by compile () if something went wrong. It sets the
+ * reg_error () is called by compile () if something went wrong. It sets the
  * status of the structure to error, and assigns the error field of the union.
  */
 #define reg_error(str)	program->status = REG_ERROR, \
 					program->result.err_mess = (str)
 
 /*
- * Bcopy copies `bytes' bytes from the `from' address into the `to' address.
+ * Bcopy copies 'bytes' bytes from the 'from' address into the 'to' address.
  */
 #ifdef vms
 #define defbcopy
@@ -893,6 +966,7 @@ char * too_long = "Regular expression too long";
 #endif
 
 #ifdef defbcopy	/* otherwise also in standard library */
+static
 void
 bcopy (from, to, bytes)
   register char * from;
@@ -928,10 +1002,10 @@ finished (program, last_exp)
 {
   register unsigned int length = (last_exp - exp_buffer) * sizeof (int);
 /* Allocate space */
-  program->result.expression = (int *) alloc (length);
+  program->result.expression = alloc (length);
   if (program->result.expression == NIL_INT) {
 	ring_bell ();
-	error ("Cannot allocate memory for search expression", NIL_PTR);
+	error ("Cannot allocate memory for search expression");
 	return ERRORS;
   }
   else {	/* Copy expression. (expression consists of ints!) */
@@ -947,8 +1021,7 @@ finished (program, last_exp)
 #define trace_indent()	trace_indentation ++;
 #define trace_undent()	trace_indentation --;
 
-static
-int trace_indentation = 0;
+static int trace_indentation = 0;
 
 static
 void
@@ -1107,7 +1180,7 @@ compile (pat_start, program)
 		    if (* pattern == '\\') {
 			pattern ++;
 		    }
-		    if (expression >= & exp_buffer [BLOCK_SIZE]) {
+		    if (expression >= & exp_buffer [MAX_EXP_SIZE]) {
 			reg_error (too_long);
 			trace_expression ((exp_buffer));
 			return FINE;
@@ -1132,9 +1205,9 @@ compile (pat_start, program)
 
 			/* Build list */
 			while (low_char <= high_char && low_char != 0) {
-				if (utf8_text == True) {
+				if (utf8_text) {
 					utflen = uniUTF (low_char, utfbuf);
-				} else if (cjk_text == True) {
+				} else if (cjk_text) {
 					utflen = cjkencode (low_char, utfbuf);
 				} else {
 					utfbuf [0] = low_char;
@@ -1142,7 +1215,7 @@ compile (pat_start, program)
 				}
 				utfpoi = utfbuf;
 				while (* utfpoi != '\0') {
-					if (expression >= & exp_buffer [BLOCK_SIZE]) {
+					if (expression >= & exp_buffer [MAX_EXP_SIZE]) {
 						reg_error (too_long);
 						trace_expression ((exp_buffer));
 						return FINE;
@@ -1155,7 +1228,7 @@ compile (pat_start, program)
 
 			advance_char ((char * *) & pattern);
 		    }
-		    if (expression >= & exp_buffer [BLOCK_SIZE - 1]) {
+		    if (expression >= & exp_buffer [MAX_EXP_SIZE - 1]) {
 			reg_error (too_long);
 			trace_expression ((exp_buffer));
 			return FINE;
@@ -1171,7 +1244,7 @@ compile (pat_start, program)
 		}
 		/* Assign negate and bracket field */
 		* acct_field |= BRACKET;
-		if (negate == True) {
+		if (negate) {
 			* acct_field |= NEGATE;
 		}
 		/*
@@ -1181,20 +1254,20 @@ compile (pat_start, program)
 		previous (expression) |= BRACKET;
 		break;
 	default :
-		if ((utf8_text == True || cjk_text == True) && multichar (c))
-		{	/* simulate single-char [u] pattern */
+		if ((utf8_text || cjk_text) && multichar (c)) {
+			/* simulate single-char [u] pattern */
 			acct_field = expression ++;
 			pattern --;	/* reset to begin of character */
 
 			/* determine multi-byte char length */
-			if (utf8_text == True) {
+			if (utf8_text) {
 				utf8_info (pattern, & utflen, & low_char);
-			} else /* (cjk_text == True) */ {
+			} else /* (cjk_text) */ {
 				utflen = CJK_len (pattern);
 			}
 
 			for (utfcount = 0; utfcount < utflen; utfcount ++) {
-				if (expression >= & exp_buffer [BLOCK_SIZE]) {
+				if (expression >= & exp_buffer [MAX_EXP_SIZE]) {
 					reg_error (too_long);
 					trace_expression ((exp_buffer));
 					return FINE;
@@ -1215,7 +1288,7 @@ compile (pat_start, program)
 			* expression ++ = c + NORMAL;
 		}
 	}
-	if (expression == & exp_buffer [BLOCK_SIZE]) {
+	if (expression == & exp_buffer [MAX_EXP_SIZE]) {
 		reg_error (too_long);
 		trace_expression ((exp_buffer));
 		return FINE;
@@ -1230,6 +1303,7 @@ compile (pat_start, program)
  * Match () will look through the whole file until a match is found.
  * NIL_LINE is returned if no match could be found.
  */
+static
 LINE *
 match (program, string, method)
   REGEX * program;
@@ -1543,7 +1617,7 @@ check_string (program, string, from_line, expression)
 	default :
 		/* panic ("Corrupted search program", NIL_PTR); */
 		ring_bell ();
-		error ("Corrupted search program", NIL_PTR);
+		error ("Corrupted search program");
 		sleep (2);
 		trace_undent ();
 		return NO_MATCH;
@@ -1642,7 +1716,7 @@ in_exprlist (exprlist, s, exprlist_length, opcode)
 	return NO_MATCH;
   }
 
-  if (utf8_text == True) {
+  if (utf8_text) {
 	while (exprlist_length > 1) {	/* > 1, don't check acct_field */
 		charmatch = True;
 		if ((* exprlist & LOW_BYTE) != * s) {
@@ -1654,7 +1728,7 @@ in_exprlist (exprlist, s, exprlist_length, opcode)
 			spoi ++;
 			while ((* exprlist & 0xC0) == 0x80) {
 				/* UTF-8 sequence byte */
-				if (charmatch == True) {
+				if (charmatch) {
 					if ((* exprlist & LOW_BYTE) != * spoi) {
 						charmatch = False;
 					}
@@ -1665,11 +1739,11 @@ in_exprlist (exprlist, s, exprlist_length, opcode)
 		} else {
 			exprlist ++; exprlist_length --;
 		}
-		if (charmatch == True) {
+		if (charmatch) {
 			return (opcode & NEGATE) ? NO_MATCH : MATCH;
 		}
 	}
-  } else if (cjk_text == True) {
+  } else if (cjk_text) {
 	while (exprlist_length > 1) {	/* > 1, don't check acct_field */
 		cjkbuf [0] = * exprlist & LOW_BYTE;
 		if (multichar (cjkbuf [0])) {
@@ -1690,7 +1764,7 @@ in_exprlist (exprlist, s, exprlist_length, opcode)
 			cjklen --;
 		}
 
-		if (charmatch == True) {
+		if (charmatch) {
 			return (opcode & NEGATE) ? NO_MATCH : MATCH;
 		}
 	}
