@@ -36,6 +36,8 @@ gettime ()
 |*			Text buffer routines				*|
 \*======================================================================*/
 
+#define dont_debug_move
+
 int old_x = 0;	/* previous x position */
 
 /*
@@ -55,7 +57,6 @@ find_x (line, address)
 	/* Alright, SHIFT_SIZE is now guaranteed to be a multiple of 8 
 	   due to lots of display problems related to that matter.
 	   Leave this code anyway. */
-  int utfcount;
   unsigned long unichar;
   int x_before = 0;
   int prevx;
@@ -70,12 +71,18 @@ find_x (line, address)
 		x_before = prevx;
 	}
   }
-  if (utf8_text == True && combining_mode == True) {
-	utf8_info (textp, & utfcount, & unichar);
+
+  /* if on combining character, skip to base character */
+  if (combining_mode && encoding_has_combining ()) {
+	unichar = unicodevalue (textp);
 	if (iscombined (unichar, textp, line->text)) {
 		x_in_line = x_before;
 	}
   }
+
+#ifdef debug_move
+  printf ("find_x %08X -> %d\n", address, x_in_line + x_left);
+#endif
   return x_in_line + x_left;
 }
 
@@ -99,14 +106,13 @@ find_address (line, new_x, cur_x)
   char * prev_poi;
   int prev_x;
   unsigned long unichar;
-  int utflen;
 
   while (tx < new_x && * textp != '\n') {
 	if (is_tab (* textp)) {
 		if (new_x == old_x /* (* cur_x) */ - 1 && tab (tx) > new_x) {
 			/* Moving left over tab */
 			break;
-		} else if (tab_left == True && new_x == old_x && tab (tx) > new_x) {
+		} else if (tab_left && new_x == old_x && tab (tx) > new_x) {
 			/* Moving up/down on tab */
 			break;
 		} else {
@@ -118,11 +124,13 @@ find_address (line, new_x, cur_x)
 		prev_x = tx;
 
 		advance_char_scr (& textp, & tx, line->text);
-		if (utf8_text == True && combining_mode == True) {
-			utf8_info (textp, & utflen, & unichar);
+
+		/* skip over combining characters */
+		if (combining_mode && encoding_has_combining ()) {
+			unichar = unicodevalue (textp);
 			while (iscombined (unichar, textp, line->text)) {
 				advance_char_scr (& textp, & tx, line->text);
-				utf8_info (textp, & utflen, & unichar);
+				unichar = unicodevalue (textp);
 			}
 		}
 
@@ -134,6 +142,9 @@ find_address (line, new_x, cur_x)
 		}
 	}
   }
+#ifdef debug_move
+  printf ("find_address %d (%d) -> %08X\n", new_x, * cur_x, textp);
+#endif
   * cur_x = tx;
   return textp;
 }
@@ -155,6 +166,7 @@ find_address (line, new_x, cur_x)
  * move_address is directly called by move_next/previous_word(), 
  * re_search(), RDwin(), load_file(), justi(), JUSandreturn()
  */
+static
 void
 move_it (new_x, new_address, new_y)
   register int new_x;
@@ -204,7 +216,7 @@ move_it (new_x, new_address, new_y)
   }
 
 /* Adjust on character boundary */
-  if (cjk_text == True) {
+  if (cjk_text) {
 	char_begin = charbegin (line->text, new_address);
 	if (char_begin != new_address) {
 		/* adjust position which is currently within a 
@@ -222,7 +234,7 @@ move_it (new_x, new_address, new_y)
 			new_address = char_begin;
 		}
 	}
-  } else if (utf8_text == True) {
+  } else if (utf8_text) {
 	if ((* new_address & 0xC0) == 0x80) {	/* UTF-8 sequence byte? */
 	/* adjust position which is currently within a UTF-8 character */
 	/* adjust new_x for multi-width characters ? */
@@ -333,8 +345,8 @@ calc_line_no ()
 }
 
 /*
- * Proceed returns the count'th line after `line'. When count is negative
- * it returns the count'th line before `line'. When the next (previous)
+ * Proceed returns the count'th line after 'line'. When count is negative
+ * it returns the count'th line before 'line'. When the next (previous)
  * line is the tail (header) indicating EOF (tof) it stops.
  */
 LINE *
@@ -355,7 +367,7 @@ proceed (line, count)
 }
 
 /*
- * Reset assigns bot_line, top_line and cur_line according to `head_line'
+ * Reset assigns bot_line, top_line and cur_line according to 'head_line'
  * which must be the first line of the screen, and a y-coordinate,
  * which will be the current y-coordinate (if it isn't larger than last_y)
  */
@@ -418,9 +430,7 @@ initialize ()
  * get_line () returns ERRORS, else it returns the length of the string.
  */
 static char * get_l_err_u1;
-static char * get_l_err_u2;
 static char * get_l_err1;
-static char * get_l_err2;
 static character last_byte = '\0';
 
 static long count_good_utf;	/* count good UTF-8 sequences */
@@ -456,9 +466,7 @@ void
 reset_get_line ()
 {
   get_l_err1 = NIL_PTR;
-  get_l_err2 = NIL_PTR;
   get_l_err_u1 = NIL_PTR;
-  get_l_err_u2 = NIL_PTR;
 
   count_good_utf = 0;
   count_bad_utf = 0;
@@ -486,26 +494,26 @@ reset_get_line ()
   Unicode_BOM = False;
 }
 
+static
 void
 show_get_l_error (get_l_err)
   char * get_l_err;
 {
   if (get_l_err != NIL_PTR) {
 	ring_bell ();
-	error (get_l_err, " -- type a blank");
+	error2 (get_l_err, " -- type a blank");
 	while (readcharacter () != ' ' && quit == False) {
 		ring_bell ();
 		flush ();
 	}
   }
 }
+
 void
 show_get_l_errors ()
 {
   show_get_l_error (get_l_err1);
-  show_get_l_error (get_l_err2);
   show_get_l_error (get_l_err_u1);
-  show_get_l_error (get_l_err_u2);
   clear_status ();
 }
 
@@ -534,7 +542,7 @@ UTF_transform (UTF8buf, maxbufl, next_byte_poi, fini_byte)
 	unichar = * * next_byte_poi;
 	(* next_byte_poi) ++;
 	if (* next_byte_poi < fini_byte) {
-		if (Unicode_little_endian == True) {
+		if (Unicode_little_endian) {
 			unichar |= (* * next_byte_poi) << 8;
 		} else {
 			unichar = (unichar << 8) | (* * next_byte_poi);
@@ -618,13 +626,13 @@ get_line (fd, buffer, len)
 	do {	/* read one effective (not to be filtered out) char */
 	    ignore1char = False;
 	    if (cur_pos == last) {
-		if (Unicode_BOM == True) {
+		if (Unicode_BOM) {
 		    if (next_byte >= fini_byte) {
 			do {
 				interrupted = False;
 				read_bytes = read (fd, UTF16buf, screen_BUFL);
 				trace_read (("read %d\n", read_bytes));
-			} while (interrupted == True ||
+			} while (interrupted ||
 				 (read_chars == -1 && geterrno () == EINTR));
 			if (read_bytes <= 0) {
 				read_chars = read_bytes;
@@ -639,7 +647,7 @@ get_line (fd, buffer, len)
 			interrupted = False;
 			read_chars = read (fd, screen, screen_BUFL);
 			trace_read (("read %d\n", read_chars));
-		    } while (interrupted == True ||
+		    } while (interrupted ||
 			     (read_chars == -1 && geterrno () == EINTR));
 		    if (read_chars <= 0) {
 			break;
@@ -653,26 +661,30 @@ get_line (fd, buffer, len)
 				/* UTF-8 BOM */
 				UTF_BOM = True;
 			} else if (strncmp (screen, "\376\377", 2) == 0
-				   && auto_detect_utf == True) {
+				   && auto_detect_utf) {
 				/* UTF-16 BOM */
 				Unicode_BOM = True;
 				Unicode_little_endian = False;
 				UTF_BOM = True;
+				/* remove BOM from text: skip UTF-8 BOM */
+				cur_pos += 3;
 			} else if (strncmp (screen, "\377\376", 2) == 0
-				   && auto_detect_utf == True) {
+				   && auto_detect_utf) {
 				/* little-endian reversed UTF-16 BOM */
 				Unicode_BOM = True;
 				Unicode_little_endian = True;
 				UTF_BOM = True;
+				/* remove BOM from text: skip UTF-8 BOM */
+				cur_pos += 3;
 			}
-			if (Unicode_BOM == True) {
-				get_l_err_u1 = "Converted UTF-16 input";
+			if (Unicode_BOM) {
+			/*	get_l_err_u0 = "Converted UTF-16 input"; */
 
-			/* move UTF-16 input to UTF-16 buffer */
+				/* move UTF-16 input to UTF-16 buffer */
 				if (UTF16buf == NIL_PTR) {
 					UTF16buf = alloc (screen_BUFL);
 					if (UTF16buf == NIL_PTR) {
-						get_l_err_u2 = "Cannot allocate memory for UTF-16 buffer";
+						get_l_err_u1 = "Cannot allocate memory for UTF-16 buffer";
 						viewonly = True;
 						modified = False;
 						return ERRORS;
@@ -683,11 +695,9 @@ get_line (fd, buffer, len)
 				next_byte = UTF16buf;
 				fini_byte = & UTF16buf [read_bytes];
 
-			/* transform to UTF-8 */
+				/* transform to UTF-8 */
 				read_chars = UTF_transform (screen, screen_BUFL, & next_byte, fini_byte);
 				last = & screen [read_chars];
-
-				modified = True;
 			}
 			count_read_op = 1;
 		}
@@ -754,7 +764,7 @@ get_line (fd, buffer, len)
 			case '':
 			case '':
 			case '':
-			case '':
+			case 0x1A:
 			case '':
 			case '':
 			case '':
@@ -841,8 +851,11 @@ get_line (fd, buffer, len)
 
 		/*	JIS	EUC-JP		A1-FE	A1-FE
 						8F	A1-FE	A1-FE
+						8E	A1-DF
 		*/
-		if (last_cjkbyte >= 0xA1 /* && last_cjkbyte <= 0xFE */
+		if (((last_cjkbyte >= 0xA1 && last_cjkbyte <= 0xFE)
+		     || last_cjkbyte == 0x8E
+		    )
 		    && curbyte >= 0xA1 && curbyte <= 0xFE)
 		{
 			count_jp ++;
@@ -931,7 +944,7 @@ get_line (fd, buffer, len)
 		ret = NUL_LINE;
 	    }
 
-	} while (ignore1char == True);
+	} while (ignore1char);
 
 	/* detect if no more lines available */
 	if (cur_pos == last) {
@@ -943,12 +956,12 @@ get_line (fd, buffer, len)
 	if (buffer > fini - 6 && * cur_pos != '\n') {
 	    /* try not to split within a UTF-8 sequence */
 	    if (buffer == fini ||	/* last chance to split! */
-		(cjk_text == True && auto_detect_utf == False
+		(cjk_text && auto_detect_utf == False
 		? charbegin (begin, buffer) == buffer
 		: (* cur_pos & 0xC0) != 0x80
 		))
 	    {
-		get_l_err2 = "Line(s) too long - split";
+		get_l_err1 = "Line(s) too long - split";
 		* buffer ++ = '\n';
 
 		/* indicate split line */
@@ -970,7 +983,7 @@ get_line (fd, buffer, len)
 
 	/* copy char from input buffer into line buffer */
 	char_copied = * buffer ++ = * cur_pos ++;
-	if (utf8_lineends == True) {
+	if (utf8_lineends) {
 		if (Ulineend_state == 0 && char_copied == '') {
 			Ulineend_state = 1;
 		} else if (Ulineend_state > 0) {
@@ -1002,7 +1015,7 @@ get_line (fd, buffer, len)
 		(* (buffer - 1) != '\r' || RET_opt != 'M')
 	   )
 	{
-		if (loading == True) {
+		if (loading) {
 		/* Add '\n' to last line of file, for internal handling */
 			* buffer ++ = '\n';
 			* buffer = '\0';
@@ -1088,7 +1101,8 @@ get_open_pos (fn)
 	(void) close (mark_fd);
 	clear_buffer ();
   }
-  modified = modif; /* don't let the line number file affect the modified flag */
+  /* prevent affecting the modified flag when loading the line number file */
+  modified = modif;
 }
 
 int
@@ -1162,13 +1176,13 @@ extract_lineend_type (text, length)
   } else if (* end == '\r') {
 	* end = '\n';
 	return lineend_CR;
-  } else if (utf8_lineends == True && length >= 3 && strncmp (end - 2, " ", 3) == 0) {
+  } else if (utf8_lineends && length >= 3 && strncmp (end - 2, " ", 3) == 0) {
 	end --;
 	end --;
 	* end ++ = '\n';
 	* end = '\0';
 	return lineend_LS;	/* Unicode line separator 2028 */
-  } else if (utf8_lineends == True && length >= 3 && strncmp (end - 2, " ", 3) == 0) {
+  } else if (utf8_lineends && length >= 3 && strncmp (end - 2, " ", 3) == 0) {
 	end --;
 	end --;
 	* end ++ = '\n';
@@ -1264,7 +1278,7 @@ load_file_position (file, with_display, to_open_linum, to_open_col)
 			get_open_pos (file_name);
 		}
 	} else {
-		error ("Cannot open: " /*, file */, serror ());
+		error2 ("Cannot open: " /*, file */, serror ());
 	}
   }
   set_file_type_flags ();
@@ -1277,7 +1291,7 @@ load_file_position (file, with_display, to_open_linum, to_open_col)
 	status_line ("Reading ", file);
 	save_utf8_text = utf8_text;
 	save_cjk_text = cjk_text;
-	if (auto_detect_utf == True) {
+	if (auto_detect_utf) {
 		utf8_text = False;
 		cjk_text = False;
 	}
@@ -1316,13 +1330,13 @@ load_file_position (file, with_display, to_open_linum, to_open_col)
 		mark_time (time_count);
 		nr_of_bytes += (long) len;
 
-		if (auto_detect_utf == True) {
+		if (auto_detect_utf) {
 			nr_of_chars += char_count (text_buffer);
 			nr_of_utfchars += utf8_count (text_buffer);
 			cjk_text = True;
 			nr_of_cjkchars += char_count (text_buffer);
 			cjk_text = False;
-		} else if (utf8_text == True) {
+		} else if (utf8_text) {
 			nr_of_chars += utf8_count (text_buffer);
 		} else {
 			nr_of_chars += char_count (text_buffer);
@@ -1395,10 +1409,10 @@ load_file_position (file, with_display, to_open_linum, to_open_col)
   }
 
   /* Unicode encoding detection */
-  if (auto_detect_utf == True) {
+  if (auto_detect_utf) {
 	mapped_text = False;
-	if (UTF_BOM == True || 
-		utf8_screen == True ? count_good_utf >= count_bad_utf
+	if (UTF_BOM || 
+		utf8_screen ? count_good_utf >= count_bad_utf
 				: count_good_utf > count_bad_utf)
 	{
 		nr_of_chars = nr_of_utfchars;
@@ -1414,7 +1428,7 @@ load_file_position (file, with_display, to_open_linum, to_open_col)
 		if (count_good_viscii / 2 > count_good_cjk + count_weak_cjk
 		    && count_good_viscii > count_good_iso * 1.2
 		   ) {
-			set_mapped_text ('V');
+			(void) set_char_encoding ('V');
 			detect_cjk = False;
 		} else if (count_good_cjk > 
 			count_bad_cjk * 2 + count_weak_cjk / 5
@@ -1433,7 +1447,7 @@ load_file_position (file, with_display, to_open_linum, to_open_col)
   /* detect CJK encoding based on 
      count_big5 count_gb count_uhc count_jp
   */
-  if (cjk_text == True && detect_cjk == True) {
+  if (cjk_text && detect_cjk) {
 	mapped_text = False;
 
 #ifdef debug_auto_detect
@@ -1451,15 +1465,15 @@ load_file_position (file, with_display, to_open_linum, to_open_col)
 #endif
 
 	if (count_gb > count_big5 && count_gb > count_uhc && count_gb > count_jp && count_gb > count_sjis) {
-		set_cjk_table ('G');
+		(void) set_char_encoding ('G');
 	} else if (count_uhc > count_big5 && count_uhc > count_gb && count_uhc > count_jp && count_uhc > count_sjis) {
-		set_cjk_table ('K');
+		(void) set_char_encoding ('K');
 	} else if (count_big5 > count_uhc && count_big5 > count_gb && count_big5 > count_jp && count_big5 > count_sjis) {
-		set_cjk_table ('B');
+		(void) set_char_encoding ('B');
 	} else if (count_jp > count_sjis) {
-		set_cjk_table ('J');
+		(void) set_char_encoding ('J');
 	} else if (count_sjis > 0) {
-		set_cjk_table ('S');
+		(void) set_char_encoding ('S');
 	}
   }
 
@@ -1499,16 +1513,16 @@ load_file_position (file, with_display, to_open_linum, to_open_col)
 	}
 	move_address (cpoi, y);
 
-	if (with_display == True) {
+	if (with_display) {
 		display (0, top_line, last_y, 0);
 	}
-	if (cjk_text == False || cjk_encoding_selected == True) {
+	if (cjk_text == False || cjk_encoding_selected) {
 		fstatus ("Read", nr_of_bytes, nr_of_chars);
 	} else {
 		fstatus ("Read", nr_of_bytes, -1L);
 	}
 	Markn (0);
-  } else if (with_display == True) {
+  } else if (with_display) {
 	display (0, top_line, last_y, 0);
 	move_to (0, 0);
   }
@@ -1540,6 +1554,7 @@ char wildfile [maxLINE_LEN];
 char drive [9], dir [maxLINE_LEN], name [13], ext [5];
 #endif
 
+static
 void
 load_wild_file (file, with_display)
   char * file;
@@ -1572,30 +1587,30 @@ load_wild_file (file, with_display)
   }
 }
 
-/*-------------------------------------------------------------------------*/
 
 /*
  * Ask user if named file should be overwritten.
  */
 FLAG
 checkoverwrite (name)
-char * name;
+  char * name;
 {
   character c;
 
   status_line ("Checking ", name);
-  if (access (name, 0 /* F_OK */) < 0)	/* Cannot access file */
+  if (access (name, 0 /* F_OK */) < 0) {	/* Cannot access file */
 	return True;	/* thus no danger of unwanted damage */
+  }
 
-  status_line (name [0] ? name : "[buffer]" ,
+  status_line (name [0] ? name : empty_buffer_name,
 			": OK to overwrite? (y/n)");
   c = promptyn ();
   clear_status ();
-  if (c == 'y')
+  if (c == 'y') {
 	return True;
-  else if (c == 'n')
+  } else if (c == 'n') {
 	return False;
-  else {
+  } else {
 /*	quit = False;	abort character has been given */
 	return False;
   }
@@ -1609,18 +1624,19 @@ NN ()
 {
   char file [maxLINE_LEN];	/* Buffer for new file name */
 
-  if (restricted == True) {
+  if (restricted) {
 	restrictederr ();
 	return;
   }
-  if (get_file ("Enter new file name:", file) == ERRORS)
+  if (get_filename ("Enter new file name:", file) == ERRORS) {
 	return;
+  }
 
   overwriteOK = False;
   writable = True;
   copy_string (file_name, file);	/* Save new file name */
 #ifdef unix
-  if (modified == True) {
+  if (modified) {
 	RD_window_title ();
   }
 #endif
@@ -1631,6 +1647,7 @@ NN ()
 
 /*
  * writechar does a buffered output to file.
+   (see also putoutchar which does the same)
  */
 int
 writechar (fd, c)
@@ -1644,6 +1661,55 @@ writechar (fd, c)
   return FINE;
 }
 
+/*
+ * writeucs writes a UCS Unicode code in UTF-16
+ * Return # words written or ERRORS.
+ */
+static
+int
+writeucs (fd, c)
+  int fd;
+  unsigned long c;
+{
+  int err = FINE;
+
+  if (c > (unsigned long) 0x10FFFF) {
+	return writeucs (fd, 0xFFFD);
+  } else if (c > (unsigned long) 0xFFFF) {
+	err = 2;
+	c -= 0x10000;
+	err |= writeucs (fd, 0xD800 | (c >> 10));
+	err |= writeucs (fd, 0xDC00 | (c & 0x03FF));
+  } else {
+	err = 1;
+	if (Unicode_little_endian) {
+		err |= writechar (fd, c & 0xFF);
+		err |= writechar (fd, c >> 8);
+	} else {
+		err |= writechar (fd, c >> 8);
+		err |= writechar (fd, c & 0xFF);
+	}
+  }
+
+  return err;
+}
+
+/*
+ * writelechar writes a line-end character to file
+ */
+static
+int
+writelechar (fd, c)
+  int fd;
+  char c;
+{
+  if (utf8_text && Unicode_BOM) {
+	return writeucs (fd, (character) c);
+  } else {
+	return writechar (fd, c);
+  }
+}
+
 int
 write_lineend (fd, return_type)
   register int fd;
@@ -1651,38 +1717,64 @@ write_lineend (fd, return_type)
 {
   switch (return_type) {
     case lineend_NONE:	return 0;
-    case lineend_NUL:	if (writechar (fd, '\0') == ERRORS)
+    case lineend_NUL:	if (writelechar (fd, '\0') == ERRORS) {
 				return ERRORS;
+			}
 			return 1;
-    case lineend_LF:	if (writechar (fd, '\n') == ERRORS)
+    case lineend_LF:	if (writelechar (fd, '\n') == ERRORS) {
 				return ERRORS;
+			}
 			return 1;
-    case lineend_CRLF:	if (writechar (fd, '\r') == ERRORS)
+    case lineend_CRLF:	if (writelechar (fd, '\r') == ERRORS) {
 				return ERRORS;
-			if (writechar (fd, '\n') == ERRORS)
+			}
+			if (writelechar (fd, '\n') == ERRORS) {
 				return ERRORS;
+			}
 			return 2;
-    case lineend_CR:	if (writechar (fd, '\r') == ERRORS)
+    case lineend_CR:	if (writelechar (fd, '\r') == ERRORS) {
 				return ERRORS;
+			}
 			return 1;
     case lineend_LS:	/* Unicode line separator 2028:   */
-			if (writechar (fd, '\342') == ERRORS)
-				return ERRORS;
-			if (writechar (fd, '\200') == ERRORS)
-				return ERRORS;
-			if (writechar (fd, '\250') == ERRORS)
-				return ERRORS;
-			return 3;
+			if (utf8_text && Unicode_BOM) {
+				if (writeucs (fd, 0x2028) == ERRORS) {
+					return ERRORS;
+				}
+				return 1;
+			} else {
+				if (writelechar (fd, '\342') == ERRORS) {
+					return ERRORS;
+				}
+				if (writelechar (fd, '\200') == ERRORS) {
+					return ERRORS;
+				}
+				if (writelechar (fd, '\250') == ERRORS) {
+					return ERRORS;
+				}
+				return 3;
+			}
     case lineend_PS:	/* Unicode paragraph separator 2029:   */
-			if (writechar (fd, '\342') == ERRORS)
-				return ERRORS;
-			if (writechar (fd, '\200') == ERRORS)
-				return ERRORS;
-			if (writechar (fd, '\251') == ERRORS)
-				return ERRORS;
-			return 3;
-    default:		if (writechar (fd, '\n') == ERRORS)
+			if (utf8_text && Unicode_BOM) {
+				if (writeucs (fd, 0x2029) == ERRORS) {
+					return ERRORS;
+				}
+				return 1;
+			} else {
+				if (writelechar (fd, '\342') == ERRORS) {
+					return ERRORS;
+				}
+				if (writelechar (fd, '\200') == ERRORS) {
+					return ERRORS;
+				}
+				if (writelechar (fd, '\251') == ERRORS) {
+					return ERRORS;
+				}
+				return 3;
+			}
+    default:		if (writelechar (fd, '\n') == ERRORS) {
 				return ERRORS;
+			}
 			return 1;
   }
 }
@@ -1690,32 +1782,57 @@ write_lineend (fd, return_type)
 /*
  * Writestring writes the given string on the given filedescriptor.
  * (buffered via writechar via misused screen buffer!)
+ * Return # bytes written or ERRORS.
  */
+static
 int
 writestring (fd, text, return_type)
-  register int fd;
-  register char * text;
+  int fd;
+  char * text;
   lineend_type return_type;
 {
-  char c;
   int len;
   int ccount = 0;
 
   while (* text != '\0') {
-	c = * text ++;
-	if (c == '\n') {
-	/* handle different line ends */
+	if (* text == '\n') {
+		/* handle different line ends */
 		len = write_lineend (fd, return_type);
-		if (len == ERRORS)
+		if (len == ERRORS) {
 			return ERRORS;
+		}
+		text ++;
 		ccount += len;
 	} else {
-		if (writechar (fd, c) == ERRORS)
-			return ERRORS;
-		ccount ++;
+		if (utf8_text && Unicode_BOM) {
+			unsigned long unichar;
+			int utflen;
+			utf8_info (text, & utflen, & unichar);
+			if (UTF_len (* text) == utflen) {
+				len = writeucs (fd, unichar);
+			} else {
+				len = writeucs (fd, 0xFFFD);
+			}
+			if (len == ERRORS) {
+				return ERRORS;
+			}
+			advance_utf8 (& text);
+			ccount += len;
+		} else {
+			if (writechar (fd, * text) == ERRORS) {
+				return ERRORS;
+			}
+			text ++;
+			ccount ++;
+		}
 	}
   }
-  return ccount;
+
+  if (utf8_text && Unicode_BOM) {
+	return ccount * 2;
+  } else {
+	return ccount;
+  }
 }
 
 /* Call graph for writing functions:
@@ -1743,18 +1860,38 @@ write_file (fd)
   int fd;
 {
   register LINE * line;
-  int ret;
+  int ret = FINE;
 
   write_count = 0L;
   chars_written = 0L;
   clear_buffer (); /* out_count = 0; */
-  for (line = header->next; line != tail; line = line->next) {
+
+  if (utf8_text && Unicode_BOM) {
+	/* for UTF-16, assure file begins with a BOM */
+	if (strncmp (header->next->text, "﻿", 3) != 0) {
+		ret = writestring (fd, "﻿", lineend_NONE);
+		if (ret == ERRORS) {
+			error2 ("Write failed (File incomplete): ", serror ());
+			write_count = -1L;
+			chars_written = -1L;
+		} else {
+			ret = FINE;
+			write_count = 2;
+			chars_written = 1;
+		}
+	}
+  }
+
+  if (ret == FINE) {
+    for (line = header->next; line != tail; line = line->next) {
 	if (line->shift_count & DUMMY) {
-		if (line->next == tail && line->text [0] == '\n')
+		if (line->next == tail && line->text [0] == '\n') {
 			continue;
+		}
 	}
 	ret = writestring (fd, line->text, line->return_type);
 	if (ret == ERRORS) {
+		error2 ("Write failed (File incomplete): ", serror ());
 		write_count = -1L;
 		chars_written = -1L;
 		break;
@@ -1764,16 +1901,22 @@ write_file (fd)
 	if (line->return_type == lineend_NONE) {
 		chars_written --;
 	}
+    }
   }
 
   if (write_count > 0L && flush_buffer (fd) == ERRORS) {
+	if (ret != ERRORS) {
+		error2 ("Write failed (File incomplete): ", serror ());
+		ret = ERRORS;
+	}
 	write_count = -1L;
 	chars_written = -1L;
   }
 
-  ret = close (fd);
-  if (ret == -1) {
-	error ("Write failed (File incomplete): ", serror ());
+  if (close (fd) == -1) {
+	if (ret != ERRORS) {
+		error2 ("Write failed (File incomplete): ", serror ());
+	}
 	write_count = -1L;
 	chars_written = -1L;
   }
@@ -1788,7 +1931,7 @@ write_text (forceflag)
   int fd;			/* Filedescriptor of file */
   int ret;
 
-  if (wpipe == True) {
+  if (wpipe) {
     fd = STD_OUT;
     status_line ("Writing ", "to standard output");
     wpipe = False; /* no further write to same stream possible */
@@ -1806,7 +1949,7 @@ write_text (forceflag)
     /* Check if file_name is valid and if file can be written */
     if (file_name [0] == '\0' || writable == False) {
 	overwriteOK = False;
-	if ((ret = get_file ("Saving edited text - enter file name:", file))
+	if ((ret = get_filename ("Saving edited text - enter file name:", file))
 		!= FINE)
 	{
 		return ret;
@@ -1814,11 +1957,12 @@ write_text (forceflag)
 	copy_string (file_name, file);		/* Save file name */
     }
     if (overwriteOK == False) {
-	if (checkoverwrite (file_name) == True) {
+	if (checkoverwrite (file_name)) {
 		overwriteOK = True;
 	} else {
-		if (quit == False)
+		if (quit == False) {
 			writable = False;
+		}
 		return ERRORS;
 	}
     }
@@ -1826,7 +1970,7 @@ write_text (forceflag)
 /*    fd = creat (file_name, fprot | xprot);	*/
     fd = open (file_name, O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, fprot | xprot);
     if (fd < 0) {	/* Empty file */
-	error ("Cannot create or write: " /*, file_name */, serror ());
+	error2 ("Cannot create or write: " /*, file_name */, serror ());
 	writable = False;
 	return ERRORS;
     } else {
@@ -1859,7 +2003,7 @@ FLAG
 save_text_load_file (fn)
   char * fn;
 {
-	if (modified == True) {
+	if (modified) {
 		if (write_text (False) == ERRORS) {
 			return ERRORS;
 		}
@@ -1875,7 +2019,7 @@ save_text_load_file (fn)
 void
 SAVEAS ()
 {
-  if (restricted == True) {
+  if (restricted) {
 	restrictederr ();
 	return;
   }
@@ -1900,7 +2044,7 @@ WT ()
 void
 WTU ()
 {
-  if (restricted == True && viewonly == True) {
+  if (restricted && viewonly) {
 	restrictederr ();
 	return;
   }
@@ -1920,12 +2064,13 @@ SAVPOS ()
 /*
  * Ask the user if he wants to save the file or not.
  */
+static
 int
 ask_save ()
 {
   register character c;
 
-  status_line (file_name [0] ? file_name : "[buffer]" ,
+  status_line (file_name [0] ? file_name : empty_buffer_name,
 			" has been modified. Save? (y/n)");
   c = promptyn ();
   clear_status ();
@@ -1960,6 +2105,7 @@ panicwrite ()
  * the file, but we would have to consider carefully the relationship 
  * between viewonly and modified.)
  */
+static
 void
 edit_file (prompt, vomode)
   char * prompt;
@@ -1967,12 +2113,12 @@ edit_file (prompt, vomode)
 {
   char new_file [maxLINE_LEN];	/* Buffer to hold new file name */
 
-  if (modified == True && viewonly == False && ask_save () != FINE) {
+  if (modified && viewonly == False && ask_save () != FINE) {
 	return;
   }
 
 /* Get new file name */
-  if (get_file (prompt, new_file) == ERRORS) {
+  if (get_filename (prompt, new_file) == ERRORS) {
 	return;
   }
 
@@ -1984,12 +2130,14 @@ edit_file (prompt, vomode)
   load_wild_file (new_file [0] == '\0' ? NIL_PTR : new_file, True);
 }
 
+#ifdef unused
+static
 void
 edit_this_file (new_file, vomode)
   char * new_file;
   FLAG vomode;
 {
-  if (modified == True && viewonly == False && ask_save () != FINE) {
+  if (modified && viewonly == False && ask_save () != FINE) {
 	return;
   }
 
@@ -2000,11 +2148,12 @@ edit_this_file (new_file, vomode)
   clearscreen ();
   load_wild_file (new_file [0] == '\0' ? NIL_PTR : new_file, True);
 }
+#endif
 
 void
 EDIT ()
 {
-  if (restricted == True) {
+  if (restricted) {
 	restrictederr ();
 	return;
   }
@@ -2014,7 +2163,7 @@ EDIT ()
 void
 VIEW ()
 {
-  if (restricted == True) {
+  if (restricted) {
 	restrictederr ();
 	return;
   }
@@ -2024,7 +2173,7 @@ VIEW ()
 void
 EDITmode ()
 {
-  if (restricted == True) {
+  if (restricted) {
 	restrictederr ();
 	return;
   }
@@ -2041,27 +2190,38 @@ VIEWmode ()
 	FSTAT ();
 	flags_changed = True;
   } else {
-	error ("Cannot view only - already modified", NIL_PTR);
+	error ("Cannot view only - already modified");
   }
 }
 
-extern FLAG viewing_help;
-static void end_view_help ();
+void
+toggleVIEWmode ()
+{
+  if (viewonly) {
+	EDITmode ();
+  } else {
+	VIEWmode ();
+  }
+}
 
+static
 void
 quit_mined ()
 {
-  if (viewing_help == True) {
+  if (viewing_help) {
 	end_view_help ();
 	return;
   }
 
   delete_yank_files ();
+
+  clear_status ();
   set_cursor (0, YMAX);
   putchar ('\n');
 #ifdef unix
   clear_window_title ();
 #endif
+  flush ();
   raw_mode (OFF);
   exit (0);
 }
@@ -2073,21 +2233,23 @@ edit_nth_file (n)
   int number;
   int index;
 
-  if (modified == True && viewonly == False && ask_save () != FINE) {
+  if (modified && viewonly == False && ask_save () != FINE) {
 	return;
   }
 
   if (n == -1) {
 	status_msg ("Edit which file (enter number) or # to reload current file");
 	index = readcharacter ();
-	if (quit == True) {
+	if (quit) {
 		return;
 	}
 	if (index == '#') {
 		n = fnami;
 	} else {
 		index = get_number ("Edit which file (enter number) ...", index, & number);
-		if (index == ERRORS) return;
+		if (index == ERRORS) {
+			return;
+		}
 		n = number - 1 + fnami_min;
 	}
   }
@@ -2126,11 +2288,11 @@ nextfile (exitiflast)
 	edit_nth_file (fnami_max);
   } else {
 #ifdef msdos
-	if (modified == True && viewonly == False && ask_save () != FINE) {
+	if (modified && viewonly == False && ask_save () != FINE) {
 		return;
 	}
 
-	if ((ffblockvalid == True) && (findnext (& ffblock)) == 0) {
+	if (ffblockvalid && findnext (& ffblock) == 0) {
 		/* Discard edit buffer, initialize and load new file */
 		initialize ();
 		clearscreen ();
@@ -2142,10 +2304,10 @@ nextfile (exitiflast)
 	ffblockvalid = False;
 #endif
 	if (fnami >= fnami_max || fnami == 0) {
-		if (exitiflast == True) {
+		if (exitiflast) {
 			quit_mined ();
 		} else {
-			error ("Already at last file", NIL_PTR);
+			error ("Already at last file");
 		}
 	} else {
 		edit_nth_file (fnami + 1);
@@ -2170,9 +2332,9 @@ PRVFILE ()
 	edit_nth_file (fnami_min);
   } else if (fnami <= fnami_min) {
 #ifdef msdos
-		error ("Already at first file name", NIL_PTR);
+		error ("Already at first file name");
 #else
-		error ("Already at first file", NIL_PTR);
+		error ("Already at first file");
 #endif
 		return;
   } else {
@@ -2192,7 +2354,7 @@ NTHFILE ()
 void
 QUED ()
 {
-  if (modified == True && viewonly == False && ask_save () != FINE) {
+  if (modified && viewonly == False && ask_save () != FINE) {
 	return;
   }
 
@@ -2206,7 +2368,7 @@ QUED ()
 void
 EXFILE ()
 {
-  if (modified == True) {
+  if (modified) {
 	if (write_text (False) != FINE) {
 		return;
 	}
@@ -2225,7 +2387,7 @@ EXFILE ()
 void
 EXMINED ()
 {
-  if (modified == True) {
+  if (modified) {
 	if (write_text (False) != FINE) {
 		return;
 	}
@@ -2240,7 +2402,7 @@ EXMINED ()
 void
 EXED ()
 {
-  if (multiexit == True) {
+  if (multiexit) {
 	EXFILE ();
   } else {
 	EXMINED ();
@@ -2253,16 +2415,16 @@ EXED ()
  * whether the file is readonly and/or modified.
 called via fstatus macro:
 fstatus (mess, bytes, chars) is
-	file_status ((mess), (bytes), file_name, \
-		     total_lines, True, writable, modified, viewonly, (chars))
+	file_status ((mess), (bytes), (chars), file_name, \
+		     total_lines, True, writable, modified, viewonly)
 also called directly from WB
 */
 void
-file_status (message, bytecount, file, lines, textstat, writefl, changed, viewing, charcount)
+file_status (message, bytecount, charcount, filename, lines, textstat, writefl, changed, viewing)
   char * message;
   register long bytecount;	/* Contains number of bytes in file */
   register long charcount;	/* Contains number of characters in file */
-  char * file;
+  char * filename;
   int lines;
   FLAG textstat;
   FLAG writefl;
@@ -2307,7 +2469,7 @@ file_status (message, bytecount, file, lines, textstat, writefl, changed, viewin
   }
 
 #ifdef buffer_info
-  if (yank_status == VALID && textstat == True) {
+  if (yank_status == VALID && textstat) {
 	/* Append buffer info */
 	/* build_string (yank_msg, " Buffer: %ld char%s.", chars_saved,
 					(chars_saved == 1L) ? "" : "s");
@@ -2321,17 +2483,17 @@ file_status (message, bytecount, file, lines, textstat, writefl, changed, viewin
   }
 #endif
 
-  if (textstat == True) {
+  if (textstat) {
 	build_string (msg,
 		(lines_per_page != 0)
 		? "%s %s%s%s%s, line %d of %d (chars/bytes: %ld/%ld), page %d"
 		: "%s %s%s%s%s, line %d of %d (chars/bytes: %ld/%ld)",
 		message,
-		(rpipe == True && * message != '[') ?
-			"standard input" : file,
-		(viewing == True) ? " (View only)" : "",
-		(changed == True) ? " (modified)" : "",
-		(writefl == False) ? " (Readonly)" : "",
+		(rpipe && * message != '[') ?
+			"standard input" : filename,
+		viewing ? " (View only)" : "",
+		changed ? " (modified)" : "",
+		writefl ? "" : " (Readonly)",
 		line_number,
 		lines,
 		charcount,
@@ -2347,11 +2509,11 @@ file_status (message, bytecount, file, lines, textstat, writefl, changed, viewin
 	build_string (msg,
 		"%s %s%s%s%s, lines: %d (chars/bytes: %ld/%ld)",
 		message,
-		(rpipe == True && * message != '[') ?
-			"standard input" : file,
-		(viewing == True) ? " (View only)" : "",
-		(changed == True) ? " (modified)" : "",
-		(writefl == False) ? " (Readonly)" : "",
+		(rpipe && * message != '[') ?
+			"standard input" : filename,
+		viewing ? " (View only)" : "",
+		changed ? " (modified)" : "",
+		writefl ? "" : " (Readonly)",
 		lines,
 		charcount,
 		bytecount
